@@ -17,7 +17,7 @@
 # endif
 #endif
 
-#ifndef	S_IFMT
+#if !defined(S_IFMT) || defined(STDC_HEADERS)
 #include <sys/stat.h>
 #endif	/* S_IFMT */
 
