@@ -26,6 +26,7 @@ extern int pathcmp();
 extern struct conscell *envarlist;
 extern struct sptree *spt_searchpath;
 extern void sp_null();
+extern void *tmalloc();
 
 /*
  * Take a path specification, e.g. ":/bin:/usr/bin" and iterate through it
@@ -83,7 +84,7 @@ path_hash(command)
 	char *command;
 {
 	char *dir, *odir, *path;
-	int n;
+	spkey_t n;
 	struct conscell *d;
 	struct spblk *spl;
 	u_int pathlen;
