@@ -59,7 +59,7 @@ extern void	sb_external();
 extern void	sb_external();
 extern void	v_set();
 extern void	v_setl();
-#ifndef	tolower
+#if	!defined(tolower) && !defined(sgi) /* XX: should use string.h or some such .. */
 extern int	tolower __((const int));
 #endif
 
