@@ -243,7 +243,8 @@ main(argc, argv)
 			killprevious(-SIGTERM, pidfile);
 		if (nrouters > 1) {
 			int pgrp, ppid = getpid();
-#ifdef	USE_BSDGETPGRP
+#ifdef	USE_BSDSETPGRP /* We assume getpgrp() and setpgrp() calling
+			  conventions match in the machine.. */
 			pgrp = getpgrp(ppid);	/* BSD  -style */
 #else
 			pgrp = getpgrp();	/* SysV/POSIX -style */
