@@ -2,19 +2,24 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  */
+/*
+ *	Lots of modifications (new guts, more or less..) by
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ */
 
 #include "hostenv.h"
 #include <stdio.h>
 #include "scheduler.h"
 #include <fcntl.h>
 #include <sys/file.h>
+#include "prototypes.h"
 
 extern int errno;
 extern time_t time();
-extern char *rendezvous;
-extern int qpctlfile();
-extern int qpchannel();
-extern int qphost();
+
+static int qpctlfile();
+static int qpchannel();
+static int qphost();
 extern char *saytime();
 
 #ifdef	USE_TIMEVAL
@@ -28,9 +33,9 @@ struct utimbuf {
 };
 #endif	/* USE_TIMEVAL */
 
-FILE *qpfp;
-char qpch;
-time_t qpnow;
+static FILE *qpfp;
+static char qpch;
+static time_t qpnow;
 
 static int qid = 0;
 
@@ -65,6 +70,8 @@ qprint(fd)
 		return;
 	}
 	qpch = '\0';
+	qid  = 0;
+
 	fprintf(qpfp, "version zmailer 1.0\nVertices:\n");
 	time(&qpnow);
 	sp_scan(qpctlfile, (struct spblk *)NULL, spt_mesh[L_CTLFILE]);
@@ -74,7 +81,12 @@ qprint(fd)
 		fprintf(qpfp, "Hosts:\n");
 		sp_scan(qphost, (struct spblk *)NULL, spt_mesh[L_HOST]);
 	}
+	fprintf(qpfp, "End:\n");
+
+	thread_report(qpfp);
+	
 	fclose(qpfp);
+
 	/* XX: I suppose we don't really need to do this if we use TCP. */
 	if (rendezvous != NULL) {
 	  time_t now = time((time_t *)0) - 1;
@@ -89,8 +101,7 @@ qprint(fd)
 	}
 }
 
-int
-qpctlfile(spl)
+static int qpctlfile(spl)
 	struct spblk *spl;
 {
 	register struct ctlfile *cfp = (struct ctlfile *)spl->data;
@@ -100,76 +111,98 @@ qpctlfile(spl)
 
 	/* assert cfp != NULL */
 	for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE]) {
-		vp->qid = ++qid;
-		fprintf(qpfp, "%d:\t%s\t%d;",
-			vp->qid, cfp->mid, vp->ngroup);
-		qpch = ' ';
-		for (i = 0; i < vp->ngroup; ++i) {
-			fprintf(qpfp, "%c%ld",
-				qpch, cfp->offset[vp->index[i]]);
-			qpch = ',';
-		}
-		if (vp->message != NULL)
-			fprintf(qpfp, "\t#%s ", vp->message);
+	  vp->qid = ++qid;
+	  fprintf(qpfp, "%d:\t%s\t%d;",
+		  vp->qid, cfp->mid, vp->ngroup);
+	  qpch = ' ';
+	  for (i = 0; i < vp->ngroup; ++i) {
+	    fprintf(qpfp, "%c%ld",
+		    qpch, cfp->offset[vp->index[i]]);
+	    qpch = ',';
+	  }
+	  if (vp->message != NULL)
+	    fprintf(qpfp, "\t#%s ", vp->message);
+	  else
+	    fprintf(qpfp, "\t#");
+
+	  i = 0;
+	  if (vp->wakeup > qpnow) {
+	    buf[0] = '\0';
+	    saytime((u_long)(vp->wakeup - qpnow), buf, 1);
+	    fprintf(qpfp, "(retry in %s", buf);
+	    ++i;
+	  } else if (vp->ce.pending) {
+	    fprintf(qpfp, "(waiting for %sslot",
+		    vp->ce.pending == SIZE_L ? "" :
+		    vp->ce.pending == L_CHANNEL ? "channel " :
+		    "thread ");
+	    ++i;
+	  } else {
+	    if (vp->proc) {
+	      fprintf(qpfp, "(running now");
+	      fprintf(qpfp," pid=%d ",vp->proc->pid);
+	      if (vp->proc->vertex == vp)
+		fprintf(qpfp,"active");
+	      else
+		if (vp->proc->vertex == NULL)
+		  fprintf(qpfp,"vtx=NULL??");
 		else
-			fprintf(qpfp, "\t#");
+		  fprintf(qpfp,"touched");
+	    } else {
+	      fprintf(qpfp, "(activation pending, thread");
+	      if (vp->thread->proc) {
+		fprintf(qpfp," pid=%d ", vp->thread->proc->pid);
+		if (vp->thread->proc->thread == vp->thread)
+		  fprintf(qpfp,"expected");
+		else
+		  fprintf(qpfp,"bygone");
+	      } else {
+		fprintf(qpfp,"fully inactive");
+	      }
+	    }
+	    ++i;
+	  }
+	  if (vp->ce.expiry > 0) {
+	    /* [mea] Want to know when it expires.. */
+	    buf[0] = '\0';
+	    saytime((long)(vp->ce.expiry - qpnow), buf, 1);
+	    fprintf(qpfp,"%sexpires in %s, tries=%d",
+		    i ? ", " : "(", buf, vp->attempts);
+	    }
+	    if (i)
+	      putc(')', qpfp);
 
-		i = 0;
-		if (vp->wakeup > qpnow) {
-			buf[0] = '\0';
-			saytime((u_long)(vp->wakeup - qpnow), buf, 1);
-			fprintf(qpfp, "(retry in %s", buf);
-			++i;
-		} else if (vp->ce.pending) {
-			fprintf(qpfp, "(waiting for %sslot",
-				vp->ce.pending == SIZE_L ? "" :
-				vp->ce.pending == L_CHANNEL ? "channel " :
-				"host ");
-			++i;
-		} else {
-			fprintf(qpfp, "(running now");
-			++i;
-		}
-		if (vp->ce.expiry > 0) {
-			/* [mea] Want to know when it expires.. */
-			buf[0] = '\0';
-			saytime((u_long)(vp->ce.expiry - qpnow),
-				buf, 1);
-			fprintf(qpfp,"%sexpires in %s",
-				i ? ", " : "(", buf);
-		}
-		if (i)
-			putc(')', qpfp);
+	    putc('\n', qpfp);
+	  }
+	  return 0;
+	}
 
-		putc('\n', qpfp);
+static int qpchannel(spl)
+	struct spblk *spl;
+{
+	register struct web *wc = (struct web *)spl->data;
+	register struct vertex *vp;
+
+	if (wc->link != NULL) {
+	  fprintf(qpfp, "%s:\t", wc->name);
+	  for (vp = wc->link; vp != NULL; vp = vp->next[L_CHANNEL])
+	    fprintf(qpfp, ">%d", vp->qid);
+	  fprintf(qpfp, "\n");
 	}
 	return 0;
 }
 
-int
-qpchannel(spl)
+static int qphost(spl)
 	struct spblk *spl;
 {
 	register struct web *wc = (struct web *)spl->data;
 	register struct vertex *vp;
 
-	(void) fprintf(qpfp, "%s:\t", wc->name);
-	for (vp = wc->link; vp != NULL; vp = vp->next[L_CHANNEL])
-		fprintf(qpfp, ">%d", vp->qid);
-	fprintf(qpfp, "\n");
-	return 0;
-}
-
-int
-qphost(spl)
-	struct spblk *spl;
-{
-	register struct web *wc = (struct web *)spl->data;
-	register struct vertex *vp;
-
-	fprintf(qpfp, "%s:\t", wc->name);
-	for (vp = wc->link; vp != NULL; vp = vp->next[L_HOST])
-		fprintf(qpfp, ">%d", vp->qid);
-	fprintf(qpfp, "\n");
+	if (wc->link != NULL) {
+	  fprintf(qpfp, "%s:\t", wc->name);
+	  for (vp = wc->link; vp != NULL; vp = vp->next[L_HOST])
+	    fprintf(qpfp, ">%d", vp->qid);
+	  fprintf(qpfp, "\n");
+	}
 	return 0;
 }
