@@ -2,6 +2,10 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  */
+/*
+ *	Lots of modifications (new guts, more or less..) by
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ */
 
 /*
  * ZMailer transport scheduler.
@@ -21,6 +25,8 @@
 #include NDIR_H
 #include "zmsignal.h"
 
+#include "prototypes.h"
+
 #ifndef	MAXNAMLEN /* POSIX.1 ... */
 #define MAXNAMLEN NAME_MAX
 #endif
@@ -48,39 +54,47 @@ char	*progname;
 char	*postoffice;
 char	*rendezvous;
 char	*pidfile = PID_SCHEDULER;
-char	*mailshare, *mailbin;
+char	*mailshare;
 char	*log;
-int	mustexit;
-int	gotalarm;
-int	dumpq;
-int	canexit;
-int	rereadcf;
-int	verbose;
-int	dlyverbose = 0;
+static int mustexit;
+static int gotalarm;
+static int dumpq;
+static int canexit;
+static int rereadcf;
+static int dlyverbose = 0;
+time_t	sched_starttime;
+int	verbose = 0;
 int	querysocket = -1;	/* fd of TCP socket to listen for queries */
-int	subsubdirs;	/* use subdirectories under each channel directory */
 int	D_alloc = 0;
-int	vtxprep_skip = 0;
-int	vtxprep_skip_any = 0;
-int	vtxprep_skip_lock = 0;
-time_t	next_dirscan	= 0;
+int	hungry_childs = 0;
+static int vtxprep_skip      = 0;
+static int vtxprep_skip_any  = 0;
+static int vtxprep_skip_lock = 0;
+static time_t next_dirscan     = 0;
+static time_t next_idlecleanup = 0;
+static struct sptree *dirscan_mesh = NULL;
 
 #include "memtypes.h"
 extern memtypes stickymem;
 
-extern SIGNAL_TYPE sig_exit(), sig_alarm(), sig_iot(),
-		   sig_readcf(), sig_chld();
+static struct ctlfile *schedule __((int fd, char *file, struct config_entry *ce, long ino));
+static struct ctlfile *vtxprep __((struct ctlfile *));
+static int  vtxfill __((struct vertex *, struct config_entry *));
+static void link_in __((int flag, struct vertex *vp, u_char *s));
+static int  lockverify __((struct ctlfile *, char *));
+static int  globmatch   __((char *, char*));
+static void vtxdo   __((struct vertex *, struct config_entry *, char *));
+
+static SIGNAL_TYPE sig_exit   __((int sig));
+static SIGNAL_TYPE sig_alarm  __((int sig));
+static SIGNAL_TYPE sig_iot    __((int sig));
+static SIGNAL_TYPE sig_readcf __((int sig));
+
 extern char *Version;
 extern char *optarg;
-extern char *progname;
-extern char *progname;
-extern char *qcf_suffix;
-extern char *qlogdir;
-extern char *qoutputfile;
 extern char *strerror __((int errno));
 extern char *strchr  /* __((char *s, char c)); */ ();
 extern char *strrchr /* __((char *s, char c)); */ ();
-extern char *vtx_to_dir __((struct vertex *vp, char *buf));
 extern int cistrcmp __((char *s1, char *s2));
 extern int efstat __((int fd, struct stat *stbuf));
 extern int emkdir __((char *dirpath, int mode));
@@ -89,35 +103,14 @@ extern int eread __((int fd, void *buf, int count));
 extern int errno;
 extern int estat __((char *filename, struct stat *stbuf));
 extern int eunlink __((char *filename));
-extern int lockaddr __((int fd, long offset, char tag1, char tag2));
-extern int loginit __((int sig));
-extern int match __((char *pattern, char *string));
-extern int numkids;
 extern int optind;
-extern int subsubdirs;
-extern int sweepinterval;
-extern struct config_entry *readconfig __((char *configfile));
-extern struct config_entry *rereadconfig __((struct config_entry *,char *));
-extern struct ctlfile *schedule __((int fd, char *file, struct config_entry *ce, long ino));
-extern struct ctlfile *slurp __((int fd, long ino)), *vtxprep __((struct ctlfile *));
+
 extern time_t time();
 extern u_char *rfc822date __((time_t *));
-extern void deletemsg __((char *, struct ctlfile *));
 extern void detach __((void));
 extern void die __((int rc, char *reason));
-extern int  doagenda __((void));
 extern void killprevious __((int signal, char *pidfile));
-extern void link_in __((int flag, struct vertex *vp, u_char *s));
-extern void mklink __((char *from, char *to));
-extern int  mux __((time_t timeout));
 extern void prversion __((char *name));
-extern void qprint __((int fd));
-extern void queryipcinit __((void));
-extern void reschedule __((struct vertex *vp, int factor, int index));
-extern void unctlfile __((struct ctlfile *cfp));
-extern void vtxdo __((struct vertex *vp, struct config_entry *ce, char *path));
-extern void vtxprint __((struct vertex *vp));
-extern void defaultconfigentry __((struct config_entry *ce, struct config_entry *));
 
 struct dirstatname {
 	struct stat st;
@@ -130,10 +123,45 @@ struct dirqueue {
 	struct dirstatname **stats;
 };
 
-extern int dirqueuescan __((const char *dir, struct dirqueue *dq));
-extern int  syncweb __((struct dirqueue *dq, struct config_entry *ce));
+static int dirqueuescan __((const char *dir, struct dirqueue *dq));
+static int syncweb __((struct dirqueue *dq, struct config_entry *ce));
 
 int global_maxkids = 1000;
+time_t now;
+FILE *statuslog = NULL;
+
+void cfp_free(cfp)
+struct ctlfile *cfp;
+{
+	struct vertex *vp, *nvp;
+	struct spblk *spl;
+
+	/* Delete from the  spt_mesh[]  */
+
+	spl = sp_lookup(cfp->id, spt_mesh[L_CTLFILE]);
+	if (spl != NULL)
+	  sp_delete(spl, spt_mesh[L_CTLFILE]);
+
+	/* And from memory */
+
+	if (cfp->head != NULL) {
+	  for (vp = cfp->head; vp != NULL; vp = nvp) {
+	    nvp = vp->next[L_CTLFILE];
+	    unvertex(vp,1); /* Don't unlink! Just free()! */
+	  }
+	} else {
+	  free(cfp->contents);
+	  free((char *)cfp);
+	}
+}
+
+static int ctl_free __((struct spblk *spl));
+static int ctl_free(spl)
+struct spblk *spl;
+{
+	cfp_free((struct ctlfile *)spl->data);
+	return 0;
+}
 
 int
 main(argc, argv)
@@ -144,14 +172,17 @@ main(argc, argv)
 	char *config, *cp;
 	int i, daemonflg, c, errflg, version, fd;
 	long offout, offerr;
-	time_t now;
 	struct config_entry *tp;
 	struct dirqueue dqb;
 	struct dirqueue *dq = &dqb;
+	int freeze = 0;
 
 	/* setlinebuf(stderr);  -- no need for this ? */
 
+	time(&sched_starttime);
+
 	memset(&dqb,0,sizeof(dqb));
+	dirscan_mesh = sp_init();
 
 	if ((progname = strrchr(argv[0], '/')) == NULL)
 		progname = argv[0];
@@ -163,19 +194,29 @@ main(argc, argv)
 	resources_maximize_nofiles();
 
 	/* The theory is, that scheduler needs circa 10 fd's for its own uses,
-	   and it will use all others on child-process return channels. */
+	   and it will use all others on child-process communication fifos. */
 
 	global_maxkids = resources_query_nofiles()-10;
 
 	postoffice = rendezvous = log = config = NULL;
 	daemonflg = 1;
 	dlyverbose = 0;
-	verbose = subsubdirs = errflg = version = 0;
-	while ((c = getopt(argc, argv, "disvf:L:N:P:Q:VW")) != EOF)
+	verbose = errflg = version = 0;
+	while ((c = getopt(argc, argv, "divf:Fl:L:N:P:Q:VW")) != EOF)
 		switch (c) {
 		case 'f':	/* override default config file */
 			config = optarg;
 			break;
+		case 'F':
+			freeze = 1;
+			break;
+		case 'l':
+			statuslog = fopen(optarg,"a");
+			if (!statuslog) {
+			  perror("Can't open statistics log file (-l)");
+			  exit(1);
+			}
+			break;
 		case 'L':	/* override default log file */
 			log = optarg;
 			break;
@@ -203,9 +244,6 @@ main(argc, argv)
 		case 'i':	/* interactive */
 			daemonflg = 0;
 			break;
-		case 's':
-			subsubdirs = 1;
-			break;
 		case 'V':
 			version = 1;
 			daemonflg = 0;
@@ -223,8 +261,6 @@ main(argc, argv)
 	}
 	if ((mailshare = getzenv("MAILSHARE")) == NULL)
 	  mailshare = MAILSHARE;
-	if ((mailbin = getzenv("MAILBIN")) == NULL)
-	  mailbin = MAILBIN;
 	if ((cp = getzenv("LOGDIR")) != NULL)
 	  qlogdir = cp;
 	if (daemonflg && log == NULL) {
@@ -243,13 +279,17 @@ main(argc, argv)
 	  setbuf(stderr,NULL);	/* [mea] No buffering on stdout.. */
 	}
 #ifdef USE_SIGREAPER
-#ifdef SIGCLD
+# ifdef SIGCLD
 	SIGNAL_HANDLE(SIGCLD,  sig_chld);
-	/* SIGNAL_HANDLE(SIGCLD,SIG_IGN);	/* Auto-reap the kids.. */
-#else
+# else
 	SIGNAL_HANDLE(SIGCHLD, sig_chld);
-	/* SIGNAL_HANDLE(SIGCHLD,SIG_IGN); */
-#endif
+# endif
+#else
+# ifdef SIGCLD
+	SIGNAL_HANDLE(SIGCLD,SIG_IGN);		/* Auto-reap the kids.. */
+# else
+	SIGNAL_HANDLE(SIGCHLD,SIG_IGN);
+# endif
 #endif
 
 #ifdef	SIGUSR1
@@ -271,7 +311,7 @@ main(argc, argv)
 	}
 	if ((tp = readconfig(config)) == NULL) {
 	  cp = emalloc(strlen(config)+50);
-	  sprintf(cp, "null control file: %s", config);
+	  sprintf(cp, "null control file, propably errors in it: %s", config);
 	  die(1, cp);
 	  /* NOTREACHED */
 	}
@@ -296,7 +336,7 @@ main(argc, argv)
 	    /* NOTREACHED */
 	  }
 	  detach();		/* leave worldy matters behind */
-	  now = time(NULL);
+	  time(&now);
 	  printf("%s: scheduler daemon (%s)\n\tstarted at %s\n",
 		 progname, Version, (char *)rfc822date(&now));
 	}
@@ -352,9 +392,8 @@ main(argc, argv)
 #ifdef	MALLOC_TRACE
 	mal_leaktrace(1);
 #endif	/* MALLOC_TRACE */
-	while (!mustexit) {
-	  int wrkcnt;
-	  time_t now, timeout;
+	do {
+	  time_t timeout;
 
 	  time(&now);
 
@@ -362,11 +401,27 @@ main(argc, argv)
 
 	  if (now >= next_dirscan) {
 	    /* Directory scan time for new jobs .. */
-	    next_dirscan = now + sweepinterval;	/* 10 seconds interval */
-	    dirqueuescan(".", dq);
+	    if (dirqueuescan(".", dq) < 150) {
+	      /* If we have more things to scan, don't quit yet! */
+	      next_dirscan = now + sweepinterval;	/* 10 seconds interval */
+	      if (dq->wrkcount > 100)
+		next_dirscan = now + 60; /* Lots of work, pick new jobs
+					    less frequently */
+	    }
 	  }
 
-	  /* Submit one item from pre-scheduler-queue into scheduler */
+	  if (now >= next_idlecleanup) {
+	    next_idlecleanup = now + 20; /* 20 second interval */
+	    idle_cleanup();
+	  }
+
+	  /* Submit one item from pre-scheduler-queue into the scheduler */
+	  if (dq->wrkcount > 100) {
+	    /* If more than 100 in queue, submit 20 in a burst.. */
+	    int i;
+	    for (i=0; i < 20; ++i)
+	      syncweb(dq, tp);
+	  }
 	  if (dq->wrkcount > 0)
 	    syncweb(dq, tp);
 
@@ -377,8 +432,8 @@ main(argc, argv)
 	  if (dq->wrkcount > 0)
 	    timeout = now;
 
-	  /* Submit possible new jobs */
-	  if (doagenda() != 0)
+	  /* Submit possible new jobs (unless frozen) */
+	  if (!freeze && doagenda() != 0)
 	    timeout = now;	/* we still have some jobs avail for start */
 
 	  gotalarm = 0;
@@ -395,20 +450,23 @@ main(argc, argv)
 	      rereadcf = 0;
 	    }
 	  } while (mux(timeout) == 0 && !gotalarm &&
-		   dq->wrkcount == 0);
-	}
+		   dq->wrkcount == 0 && !mustexit);
+	} while (!mustexit);
+
+	/* Doing nicely we would kill the childs here, but we are not
+	   such a nice people -- we just discard incore data, and crash out.. */
+
+	sp_scan(ctl_free, NULL, spt_mesh[L_CTLFILE]);
+
 #ifdef	MALLOC_TRACE
 	mal_dumpleaktrace(stderr);
 #endif	/* MALLOC_TRACE */
 	if (mustexit)
 		die(0, "signal");
-	exit(0);
-	/* NOTREACHED */
 	return 0;
 }
 
-SIGNAL_TYPE
-sig_exit(sig)
+static SIGNAL_TYPE sig_exit(sig)
 int sig;
 {
 	
@@ -421,8 +479,7 @@ int sig;
 	mustexit = 1;
 }
 
-SIGNAL_TYPE
-sig_alarm(sig)
+static SIGNAL_TYPE sig_alarm(sig)
 int sig;
 {
 	gotalarm = 1;
@@ -430,8 +487,7 @@ int sig;
 }
 
 #ifdef SIGUSR2
-SIGNAL_TYPE
-sig_iot(sig)
+static SIGNAL_TYPE sig_iot(sig)
 int sig;
 {
 	dumpq = 1;
@@ -440,8 +496,7 @@ int sig;
 #endif
 
 #ifdef SIGUSR1
-SIGNAL_TYPE
-sig_readcf(sig)
+static SIGNAL_TYPE sig_readcf(sig)
 int sig;
 {
 	rereadcf = 1;
@@ -453,8 +508,7 @@ int sig;
  * Absorb any new files that showed up in our directory.
  */
 
-static int
-in_dirscanqueue(dq,ino)
+static int in_dirscanqueue(dq,ino)
 	struct dirqueue *dq;
 	long ino;
 {
@@ -462,16 +516,12 @@ in_dirscanqueue(dq,ino)
 
 	/* Return 1, if can find the "ino" in the queue */
 
-	if (dq->wrkcount == 0) return 0;
-	for (i=0; i < dq->wrkcount; ++i)
-	  if (ino == dq->stats[i]->st.st_ino)
-	    return 1;
-
+	if (dq->wrkcount == 0 || dirscan_mesh == NULL) return 0;
+	if (sp_lookup(ino,dirscan_mesh) != NULL) return 1;
 	return 0;
 }
 
-static int
-dq_ctimecompare(a,b)
+static int dq_ctimecompare(a,b)
 const void *a;
 const void *b;
 {
@@ -503,11 +553,16 @@ static int dirqueuescan(dir, dq)
 	modtime = stbuf.st_mtime;
 #endif
 
+if (verbose) { printf("dirqueuescan(dir='%s') ",dir); fflush(stdout); }
+
 	/* Some changes lately, open the dir and read it */
 
 	dirp = opendir(dir);
 	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
 	  /* Scan filenames into memory */
+
+	  if (newents > 200) break; /* At most 200 per one go */
+
 	  if (dp->d_name[0] >= '0' &&
 	      dp->d_name[0] <= '9') {
 	    /* A file whose name STARTS with a number (digit) */
@@ -544,6 +599,7 @@ static int dirqueuescan(dir, dq)
 	    memcpy(&(dsn->st),&stbuf,sizeof(stbuf));
 	    dsn->ino = ino;
 	    strcpy(dsn->name,file);
+	    sp_install(ino,(u_char*)dsn,0,dirscan_mesh);
 
 	    if (dq->wrkspace <= dq->wrkcount) {
 	      /* Increase the space */
@@ -579,11 +635,12 @@ static int dirqueuescan(dir, dq)
 #endif
 	closedir(dirp);
 
+	if (verbose) printf("wrkcount=%d new=%d\n",dq->wrkcount,newents);
+
 	return newents;
 }
 
-int
-syncweb(dq, cehead)
+static int syncweb(dq, cehead)
 	struct dirqueue *dq;
 	struct config_entry *cehead;
 {
@@ -596,6 +653,9 @@ syncweb(dq, cehead)
 	/* Any work to do ? */
 	if (dq->wrkcount == 0) return 0;
 
+	/* Be responsive, check query channel */
+	queryipccheck();
+
 	/* Ok some, decrement the count to change it to index */
 	dq->wrkcount -= 1;
 	file  =   dq->stats[dq->wrkcount]->name;
@@ -603,6 +663,8 @@ syncweb(dq, cehead)
 	ino   =   dq->stats[dq->wrkcount]->ino;
 	/* Now we have pointers */
 
+	sp_delete(sp_lookup(ino,dirscan_mesh),dirscan_mesh); /* XX: should ALWAYS succeed.. */
+
 	/* Sometimes we may get files already in processing
 	   into our pre-schedule queue */
 
@@ -638,8 +700,7 @@ syncweb(dq, cehead)
  * of transport directives and schedules the appropriate things to be
  * done at the appropriate time in the future.
  */
-struct ctlfile *
-schedule(fd, file, cehead, ino)
+static struct ctlfile *schedule(fd, file, cehead, ino)
 	int fd;
 	char *file;
 	struct config_entry *cehead;
@@ -654,12 +715,13 @@ schedule(fd, file, cehead, ino)
 	    eunlink(file);	/* everything here has been processed */
 	    if (verbose)
 	      printf("completed, unlink %s\n",file);
+	    return NULL;
 	  }
 	  vtxprep_skip_any += vtxprep_skip;
 	  return NULL;
 	}
 	if (cfp->head == NULL) {
-	  unctlfile(cfp);
+	  unctlfile(cfp, 0);
 	  return NULL;
 	}
 	for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE]) {
@@ -670,6 +732,11 @@ schedule(fd, file, cehead, ino)
 	return cfp;
 }
 
+/*
+ * slurp() gets in the job-descriptor file, and does initial
+ *         parsing on it.
+ */
+
 struct ctlfile *
 slurp(fd, ino)
 	int fd;
@@ -706,8 +773,8 @@ slurp(fd, ino)
 	cfp->fd = fd;
 	cfp->vfp = NULL;
 	cfp->head = NULL;
-	cfp->atspool = NULL;
 	cfp->uid = stbuf.st_uid;
+	cfp->envctime = stbuf.st_ctime;
 	cfp->mark = V_NONE;
 	cfp->haderror = 0;
 	cfp->contents = contents;
@@ -763,7 +830,7 @@ slurp(fd, ino)
 	/* copy over the offsets */
 	while (--i >= 0)
 	  *lp++ = *ip++;
-	cfp->id  = ino;
+	cfp->id = ino;
 	cfp->mid = NULL;
 	free(offset);	/* release the block */
 	return cfp;
@@ -777,12 +844,13 @@ struct offsort {
 	char	*sender;
 };
 
-/* Used by the qsort comparison routine, bcp points to the control file bytes */
+/* ``bcfcn'' is used by the qsort comparison routine,
+   bcp points to the control file bytes
+ */
+
 static char *bcp;
 
-int
-bcfcn(a, b)
-	struct offsort *a, *b;
+static int bcfcn(a, b) struct offsort *a, *b;
 {
 	return cistrcmp(bcp + a->offset, bcp + b->offset);
 }
@@ -814,8 +882,18 @@ lockverify(cfp,cp)		/* Return 1 when lock process does not exist */
 	return 0;	/* Locking PID does exist.. */
 }
 
-struct ctlfile *
-vtxprep(cfp)
+/*
+ *  The  vtxprep() does deeper analysis on jobs described at the file.
+ *  It verifies possible locks (if they are still valid), and gathers
+ *  all of the information regarding senders and recipients.
+ *
+ *  All "recipient"-lines are sorted to ease searching of vertices with
+ *  identical channel, and host definitions.  If there are more than one
+ *  recipient with given (channel, host)-tuple, all such recipients are
+ *  wrapped into same vertex node with its respective ``recipient group''.
+ *
+ */
+static struct ctlfile *vtxprep(cfp)
 	struct ctlfile *cfp;
 {
 	register int i, opcnt;
@@ -827,7 +905,7 @@ vtxprep(cfp)
 	struct spblk *spl;
 	struct vertex *vp, *pvp, **pvpp, *head;
 	struct stat stbuf;
-	struct offsort *offarr = NULL;
+	struct offsort *offarr;
 	int offspc;
 
 	if (cfp == NULL)
@@ -835,9 +913,11 @@ vtxprep(cfp)
 
 	sp_install(cfp->id, (u_char *)cfp, 0, spt_mesh[L_CTLFILE]);
 	channel = host = NULL;
+
 	/* copy offsets into local array */
 	offspc = 16;
 	offarr = (struct offsort *)emalloc(offspc * sizeof(struct offsort));
+
 	opcnt = 0;
 	svn = 0;
 	vtxprep_skip = 0;
@@ -846,11 +926,8 @@ vtxprep(cfp)
 	  cp = cfp->contents + *lp + 1;
 	  if (*cp == _CFTAG_LOCK) {
 	    /*
-	     * This is assumed to happen when we move old control
-	     * files into the scheduler directory. Since this is
-	     * the only processing that needs to be done except
-	     * for the rename(), we might as well do it here,
-	     * instead of in some separate program.
+	     * This can happen when we restart the scheduler, and
+	     * some previous transporter is still running.
 	     *
 	     */
 	    if (!lockverify(cfp,cp)) {
@@ -860,18 +937,15 @@ vtxprep(cfp)
 	       * won't errorneously delete it.
 	       */
 	      close(cfp->fd);	/* Was opened on  schedule() */
-	      spl = sp_lookup(cfp->id, spt_mesh[L_CTLFILE]);
-	      if (spl != NULL)
-		sp_delete(spl, spt_mesh[L_CTLFILE]);
 	      if (cfp->vfp != NULL) {
 		fprintf(cfp->vfp,
 			"New scheduler: Skipped a job-file because it is held locked by PID=%6.6s\n",cp+1);
 		fclose(cfp->vfp);
 	      }
-	      free(cfp->contents);
-	      free((char *)cfp);
+	      cfp_free(cfp);
 	      ++vtxprep_skip;
 	      ++vtxprep_skip_lock;
+	      free(offarr);
 	      return NULL;
 	    }
 	    *cp = _CFTAG_NORMAL; /* unlock it */
@@ -888,7 +962,7 @@ vtxprep(cfp)
 	      while (*cp != 0 && *cp != ' ') ++cp; while (*cp == ' ') ++cp;
 	      while (*cp != 0 && *cp != ' ') ++cp; while (*cp == ' ') ++cp;
 	      latest_sender = cp;
-	      while (*cp != 0 && *cp != ' ') ++cp;  *cp = 0; /* No spaces at the sender addr! */
+	      while (*cp != 0 && *cp != ' ') ++cp;  *cp = 0; /* We assume: No spaces at the sender addr! */
 	      cfp->erroraddr = latest_sender;
 	      break;
 	    case _CF_RECIPIENT:
@@ -1026,16 +1100,13 @@ vtxprep(cfp)
 	  sprintf(mfpath, "../%s/%s", QUEUEDIR, cfp->mid);
 	if (cfp->mid == NULL || cfp->logident == NULL
 	    || estat(mfpath, &stbuf) < 0) {
-	  spl = sp_lookup(cfp->id, spt_mesh[L_CTLFILE]);
-	  if (spl != NULL)
-	    sp_delete(spl, spt_mesh[L_CTLFILE]);
 	  if (cfp->vfp != NULL) {
 	    fprintf(cfp->vfp,
 		    "aborted due to missing information\n");
 	    fclose(cfp->vfp);
 	  }
-	  free(cfp->contents);
-	  free((char *)cfp);
+	  cfp_free(cfp);
+	  free(offarr);
 	  return NULL;
 	}
 	cfp->ctime = stbuf.st_ctime;
@@ -1060,12 +1131,11 @@ vtxprep(cfp)
 	    continue;
 	  *echannel = '\0';
 	  host = echannel + 1;
-	  while (isascii(*host) && isspace(*host))
-	    ++host;
+	  while (*host == ' ') ++host;
 #if 1
 	  /* [mea]   channel ((mx.target)(mx.target mx.target)) rest.. */
 	  cp = host;
-	  if (*cp == ')') {
+	  if (*cp == '(') {
 	    while(*cp == '(')
 	      ++cp;
 	    while(*cp && *cp != ' ' && *cp != '\t' && *cp != ')')
@@ -1104,24 +1174,23 @@ vtxprep(cfp)
 					  (i - svn - 1) * sizeof (int));
 	      vp = (struct vertex *)emalloc(alloc_size);
 	      memset((char*)vp, 0, alloc_size);
-	      vp->cfp = cfp;
-	      vp->linkcnt  = 0;
-	      vp->linkpath = NULL;
+	      vp->cfp          = cfp;
+	      vp->linkcnt      = 0;
 	      defaultconfigentry(&vp->ce,NULL);
 	      vp->next[L_CTLFILE] = NULL;
 	      vp->prev[L_CTLFILE] = pvp;
-	      vp->message = NULL;
-	      vp->retryindex = 0;
-	      vp->nextitem = NULL;
-	      vp->previtem = NULL;
-	      vp->wakeup = 0;
-	      vp->proc = NULL;
-	      vp->ngroup = i - svn;
-	      vp->attempts = 0;
-	      vp->notary = NULL;
-	      vp->sender = offarr[i].sender;
-	      vp->headeroffset = offarr[i].headeroffset; /* Just any of them will do */
-	      vp->drptoffset = offarr[i].drptoffset;
+	      vp->message      = NULL;
+	      vp->retryindex   = 0;
+	      vp->nextitem     = NULL;
+	      vp->previtem     = NULL;
+	      vp->wakeup       = 0;
+	      vp->proc         = NULL;
+	      vp->ngroup       = i - svn;
+	      vp->attempts     = 0;
+	      vp->notary       = NULL;
+	      vp->sender       = offarr[svn].sender;
+	      vp->headeroffset = offarr[svn].headeroffset; /*They are similar*/
+	      vp->drptoffset   = offarr[svn].drptoffset;
 	      while (svn < i) {
 		vp->index[i-svn-1] = offarr[svn].myidx;
 		++svn;
@@ -1129,20 +1198,21 @@ vtxprep(cfp)
 	      *pvpp = vp;
 	      pvpp = &vp->next[L_CTLFILE];
 	      pvp = vp;
-	      link_in(L_HOST, vp, (u_char *)l_host);
+	      link_in(L_HOST,    vp, (u_char *)l_host);
 	      link_in(L_CHANNEL, vp, (u_char *)l_channel);
 	    }
 	    /* create a new vertex node */
 	    svn = i;
 	  }
-	  /* stick the current 'r'eceived line into the current vertex */
+	  /* stick the current 'r'ecipient  line into the current vertex */
 	  /* restore the characters */
 	  *l_echannel = *l_ehost = ' ';
-	  l_echannel = echannel;
-	  l_ehost = ehost;
-	  l_channel = channel;
-	  l_host = host;
-	}
+	  l_echannel  = echannel;
+	  l_ehost     = ehost;
+	  l_channel   = channel;
+	  l_host      = host;
+	} /* for( .. i < opcnt .. ) */
+
 	/* wrap and tie the old vertex node (this is a copy of code above) */
 	if (i != svn) {
 	  u_int alloc_size = (u_int) (sizeof (struct vertex) +
@@ -1151,7 +1221,6 @@ vtxprep(cfp)
 	  memset((void*)vp, 0, alloc_size);
 	  vp->cfp = cfp;
 	  vp->linkcnt  = 0;
-	  vp->linkpath = NULL;
 	  defaultconfigentry(&vp->ce,NULL);
 	  vp->next[L_CTLFILE] = NULL;
 	  vp->prev[L_CTLFILE] = pvp;
@@ -1177,6 +1246,7 @@ vtxprep(cfp)
 	  link_in(L_HOST, vp, (u_char *)host);
 	  link_in(L_CHANNEL, vp, (u_char *)channel);
 	}
+
 	*l_echannel = *l_ehost = ' ';
 	/*
 	   for (vp = head; vp != NULL; vp = vp->next[L_CTLFILE]) {
@@ -1186,26 +1256,31 @@ vtxprep(cfp)
 	   }
 	*/
 	cfp->head = head;
+	free(offarr);
 	return cfp;
 }
 
-int
-vtxfill(vp, tp)
+/*
+ *  The  vtxfill()  is a subroutine to  vtxdo()  taking care of filling in
+ *  scheduler definition entries from scheduler configuration table.
+ *
+ */
+
+static int vtxfill(vp, tp)
 	struct vertex *vp;
 	struct config_entry *tp;
 {
 	/* if the channel doesn't match, there's no hope! */
 	if (verbose>1)
-	  printf("ch? %s %s\n",
-		 vp->orig[L_CHANNEL]->name, tp->channel);
+	  printf("ch? %s %s\n", vp->orig[L_CHANNEL]->name, tp->channel);
 	if (tp->channel[0] == '*' && tp->channel[1] == '\0')
 	  return 0; /* Never match the defaults entry! */
-	if (!match(tp->channel, vp->orig[L_CHANNEL]->name))
+	if (!globmatch(tp->channel, vp->orig[L_CHANNEL]->name))
 	  return 0;
 
-	if (!(tp->host == NULL || tp->host[0] == '\0'
-	      || (tp->host[0] == '*' && tp->host[1] == '\0'))) {
-	  if (!match(tp->host, vp->orig[L_HOST]->name))
+	if (!(tp->host == NULL || tp->host[0] == '\0' ||
+	      (tp->host[0] == '*' && tp->host[1] == '\0'))) {
+	  if (!globmatch(tp->host, vp->orig[L_HOST]->name))
 	    return 0;
 	}
 
@@ -1213,34 +1288,48 @@ vtxfill(vp, tp)
 	  printf("host %s %s\n", vp->orig[L_HOST]->name, tp->host);
 
 	if (tp->interval != -1) vp->ce.interval = tp->interval;
-	if (tp->expiry   != -1) vp->ce.expiry = tp->expiry;
+	if (tp->idlemax  != -1) vp->ce.idlemax  = tp->idlemax;
+	if (tp->expiry   != -1) vp->ce.expiry   = tp->expiry;
 	vp->ce.expiryform     = tp->expiryform;
 	vp->cfp->deliveryform = tp->deliveryform;
 	if (tp->uid      != -1) vp->ce.uid = tp->uid;
 	if (tp->gid      != -1) vp->ce.gid = tp->gid;
 	if (tp->maxkids  != -1) vp->ce.maxkids = tp->maxkids;
 	if (tp->maxkidChannel != -1) vp->ce.maxkidChannel = tp->maxkidChannel;
-	if (tp->maxkidHost != -1) vp->ce.maxkidHost = tp->maxkidHost;
+	if (tp->maxkidThreads != -1) vp->ce.maxkidThreads = tp->maxkidThreads;
 	if (tp->nretries > 0) {
 	  vp->ce.nretries = tp->nretries;
-	  vp->ce.retries = tp->retries;
+	  vp->ce.retries  = tp->retries;
 	}
 	if (tp->command != NULL) {
 	  vp->ce.command = tp->command;
-	  vp->ce.argv = tp->argv;
+	  vp->ce.argv    = tp->argv;
 	}
 	if (tp->bychannel != -1)
 	  vp->ce.bychannel = tp->bychannel;
-	if (tp->byhost != -1)
-	  vp->ce.byhost = tp->byhost;
+	else
+	  vp->ce.bychannel = 0;
+	if (tp->withhost != -1)
+	  vp->ce.withhost = tp->withhost;
+	else
+	  vp->ce.withhost = 0;
 	vp->ce.flags |= tp->flags;
-	vp->ce.mark = vp->ce.pending = 0;
+	vp->ce.mark   = vp->ce.pending = 0;
+	vp->ce.host   = tp->host;
 	if (tp->skew > 0) vp->ce.skew = tp->skew;
 	return 1;
 }
 
-void
-vtxdo(vp, cehead, path)
+/*
+ *  The  vtxdo()  tries thru all scheduler configuration entries
+ *  looking for a matching one which to fill in for the input vertex.
+ *
+ *  In the end it calls  reschedule()  to place the vertex into scheduling
+ *  queues (threads)
+ *
+ */
+
+static void vtxdo(vp, cehead, path)
 	struct vertex *vp;
 	struct config_entry *cehead;
 	char *path;
@@ -1261,10 +1350,8 @@ vtxdo(vp, cehead, path)
 	    break;
 	}
 	if (n == 0) {
-	  fprintf(stderr,
-		  "%s: no pattern matched %s/%s address\n",
-		  progname,
-		  vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name);
+	  fprintf(stderr, "%s: no pattern matched %s/%s address\n",
+		  progname, vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name);
 	  /* XX: memory leak here? */
 	  return;
 	}
@@ -1272,7 +1359,6 @@ vtxdo(vp, cehead, path)
 	  printf("Matched %dth config entry with: %s/%s\n", cnt,
 		 vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name);
 
-
 	/* set default values */
 	if (vp->ce.expiry > 0)
 	  vp->ce.expiry += vp->cfp->ctime;
@@ -1280,50 +1366,64 @@ vtxdo(vp, cehead, path)
 	  vp->ce.expiry = 0;
 	if (vp->ce.interval == -1)
 	  vp->ce.interval = 3600;
+	if (vp->ce.idlemax == -1)
+	  vp->ce.idlemax = vp->ce.interval * 3;
 	if (vp->ce.maxkids == -1)
 	  vp->ce.maxkids = global_maxkids;
 	if (vp->ce.maxkidChannel == -1)
 	  vp->ce.maxkidChannel = global_maxkids;
-	if (vp->ce.maxkidHost == -1)
-	  vp->ce.maxkidHost = global_maxkids;
-	if (vp->ce.byhost == -1)
-	  vp->ce.byhost = 0;
+	if (vp->ce.maxkidThreads == -1)
+	  vp->ce.maxkidThreads = global_maxkids;
+	if (vp->ce.withhost == -1)
+	  vp->ce.withhost = 0;
 	if (vp->ce.bychannel == -1)
 	  vp->ce.bychannel = 0;
 
+	thread_linkin(vp,tp,cnt);
+
 	if (verbose>1)
 	  vtxprint(vp);
-
-	if (vp->linkpath != NULL)
-	  free(vp->linkpath);
-	vp->linkpath = strsave(path);
-	if ((vp->ce.flags & FF_GANGSCHEDULE)
-	    && vp->prev[L_HOST] != NULL
-	    && vp->prev[L_CHANNEL] != NULL
-	    && vp->ce.retries == vp->prev[L_HOST]->ce.retries
-	    && vp->orig[L_HOST] == vp->prev[L_HOST]->orig[L_HOST]
-	    && vp->orig[L_CHANNEL] == vp->prev[L_CHANNEL]->orig[L_CHANNEL]) {
-	  vp->retryindex = vp->prev[L_HOST]->retryindex;
-	  reschedule(vp, -(vp->prev[L_HOST]->wakeup), -1);
-	} else
-	  reschedule(vp, 0, -1);
 }
 
+
+int vtxredo(spl)
+        struct spblk *spl;
+{
+        struct ctlfile *cfp = (struct ctlfile *)spl->data;
+	struct vertex *vp;
+
+        /* assert cfp != NULL */
+	for (vp = cfp->head ; vp != NULL ; vp = vp->next[L_CTLFILE]) {
+		vp->ce.command = NULL;
+		vp->ce.argv = NULL;
+		vp->ce.retries = NULL;
+		vp->ce.nretries = 0;
+		vtxdo(vp, rrcf_head, NULL);
+	}
+        return 0;
+}
+
+
 /* Shell-GLOB-style matching */
-int
-match(pattern, string)
+static int globmatch(pattern, string)
 	register char	*pattern;
 	register char	*string;
 {
 	while (1) {
 	  switch (*pattern) {
 	  case '*':
-	    pattern++;
+	    ++pattern;
 	    do {
-	      if (match(pattern, string))
+	      if (globmatch(pattern, string))
 		return 1;
 	    } while (*string++ != '\0');
 	    return 0;
+	  case '\\':
+	    ++pattern;
+	    if (*pattern == 0 ||
+		*pattern != *string)
+	      return 0;
+	    break;
 	  case '[':
 	    if (*string == '\0')
 	      return 0;
@@ -1349,7 +1449,7 @@ match(pattern, string)
 	    string++;
 	    break;
 	  case '?':
-	    pattern++;
+	    ++pattern;
 	    if (*string++ == '\0')
 	      return 0;
 	    break;
@@ -1369,31 +1469,15 @@ match(pattern, string)
  * happens we need 2 (host and channel names), and we don't care what
  * they are as long as they are in separate spaces.
  */
-void
-link_in(flag, vp, s)
+static void link_in(flag, vp, s)
 	int flag;
 	struct vertex *vp;
 	u_char *s;
 {
-	register u_char *cp;
-	struct web *wp;
-	struct spblk *spl;
+	struct web *wp = web_findcreate(flag,s);
+
+	wp->linkcnt += 1;
 
-	for (cp = s; *cp != '\0'; ++cp)
-	  if (isascii(*cp) && isupper(*cp))
-	    *cp = tolower(*cp);
-	spl = sp_lookup(symbol(s), spt_mesh[flag]);
-	if (spl == NULL || (wp = (struct web *)spl->data) == NULL) {
-	  wp = (struct web *)emalloc(sizeof (struct web));
-	  memset((void*)wp, 0, sizeof (struct web));
-	  sp_install(symbol(s), (u_char *)wp, 0, spt_mesh[flag]);
-	  wp->name = strsave((char *)s);
-	  wp->kids = 0;
-	  wp->link = NULL;
-	  wp->lastlink = NULL;
-	  wp->fifohead = NULL;
-	  wp->fifotail = NULL;
-	}
 	vp->next[flag] = NULL;
 	vp->prev[flag] = wp->lastlink;
 	vp->orig[flag] = wp;
@@ -1404,42 +1488,19 @@ link_in(flag, vp, s)
 	  wp->link = vp;
 }
 
-
 /*
- * Generate a file name corresponding to a link to the control file.
- * On busy systems, this should probably map into a channel/host/file
- * pathname. Unfortunately, some OS's cannot handle long host names.
- * If that becomes a problem, it should be possible to select another
- * scheme on a per-channel basis. It should be kept in mind that any
- * scheme which produces an N-to-M reverse mapping (if the host name
- * portion of the pathname is truncated or missing, say), requires
- * some smarts elsewhere to weed out duplicate link pathnames.
+ * time-string to be prepended on logged messages
  */
 char *
-vtx_to_dir(vp, buf)
-	struct vertex *vp;
-	char *buf;
+timestring()
 {
-	strcpy(buf, vp->orig[L_CHANNEL]->name);
-	strcat(buf, "/");
-	strcat(buf, vp->orig[L_HOST]->name);
-	return buf;
-}
+	static char timebuf[40];
+	struct tm *tp;
 
-struct ctlfile *
-file_to_ctl(s)
-	char *s;
-{
-	char *file;
-	struct spblk *spl;
-
-	if ((file = strrchr(s, '/')) == NULL)
-	  file = s;
-	else
-	  ++file;
-
-	spl = sp_lookup((u_long)atol(file), spt_mesh[L_CTLFILE]);
-	if (spl == NULL)
-	  return NULL;
-	return (struct ctlfile *)spl->data;
+	time(&now);
+	tp = localtime(&now);
+	sprintf(timebuf,"%d%02d%02d%02d%02d%02d",
+		tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday,
+		tp->tm_hour, tp->tm_min, tp->tm_sec);
+	return timebuf;
 }
