@@ -2,6 +2,11 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  */
+/*
+ *	Lots of modifications (new guts, more or less..) by
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ */
+
 
 #include <stdio.h>
 #include <sys/param.h>
@@ -14,14 +19,16 @@
 #include <sys/file.h>
 #include NDIR_H
 #include "zmsignal.h"
+#include <stdlib.h>
+#include <unistd.h>
 
 #include "prototypes.h"
 
-static int scheduler_nofiles = -1; /* Will be filled below */
-static void dotransport __((char **, int, struct vertex *, char*, char*));
-static void runcommand __((char **, int, char **, struct vertex *, struct web *, struct web *));
-static void stashprocess __((int, int, int, struct web *, struct web *, struct vertex *));
-static void reclaim __((int));
+
+static int  scheduler_nofiles = -1; /* Will be filled below */
+static int  runcommand __((char **, struct vertex *, struct web *, struct web*));
+static void stashprocess __((int, int, int, struct web*, struct web*, struct vertex *));
+static void reclaim __((int, int));
 static void waitandclose __((int));
 static void readfrom __((int));
 
@@ -41,97 +48,73 @@ int	numkids = 0;
 extern char *getzenv();
 extern char *strchr();
 extern int cistrcmp(), ranny();
-extern int epipe();
 extern int errno;
 extern time_t time();
 
 
-/*
- * Collect up all the marked control file names (unmarking them in the process)
- * and send them to the command given by the parameter. Said command must be
- * forked off, and the appropriate structures set up to deal with the stdout
- * of that command.
- */
-
-static int t_n, t_nmax;
-static char **t_files;
-
-static int tscan(spl)
-	struct spblk *spl;
-{
-	struct ctlfile *cfp = (struct ctlfile *)spl->data;
-
-	/* assert cfp != NULL */
-
-	/* Is this marked ? */
-	if (cfp->mark == V_NONE)
-	  return -1; /* .. No. */
-
-	/* If the count of jobs exceeds our high guesses.. */
-	if (t_n >= t_nmax) {
-	  /* Report only once per scan.. */
-	  if (t_nmax > 0)
-	    printf("**AARGH**  MAXFILESPERTRANSPORT(%d) TURNED OUT TO BE TOO SMALL!\n", t_nmax);
-	  t_nmax = 0;
-	  /* Clear the markings.. */
-	  cfp->mark    = V_NONE;
-	  return -1;
-	}
-
-	t_files[t_n++] = cfp->mid;
-	cfp->mark    = V_NONE;
-
-	return 0;
-}
 
 void
-transport(vhead, channel, host)
-	struct vertex *vhead;
-	char *channel, *host;
+feed_child(proc)
+struct procinfo *proc;
 {
-	int i;
-	char *files[MAXFILESPERTRANSPORT];	/* XX */
+	char buf[500]; /* 60 or so SHOULD suffice.. */
+	int len, rc;
 
-	/* generate the file names we care about */
-	t_n     = 0;
-	t_nmax  = MAXFILESPERTRANSPORT;
-	t_files = files;
+	if (proc->thread == NULL) return; /* Might be called without next process.. */
+
+	if (proc->hungry) --hungry_childs;
+	/* It will be fed, clear this flag.. */
+	proc->hungry = 0;
+
+	proc->vertex->proc = proc; /* Flag that it is in processing */
+
+	if (proc->thg->withhost)
+	  sprintf(buf, "%s\n", proc->vertex->cfp->mid);
+	else
+	  sprintf(buf, "%s\t%s\n", proc->vertex->cfp->mid, proc->ho->name);
+
+	len = strlen(buf);
 
 	if (verbose)
-	  printf("transport(vhead,chan=\"%s\",host=\"%s\")\n",channel,host);
+	  printf("feed: tofd=%d, len=%d, chan=%s, buf=%s",
+		 proc->tofd,len,proc->ch->name,buf);
 
-	sp_scan(tscan, (struct spblk *)NULL, spt_mesh[L_CTLFILE]);
-	dotransport(files, t_n, vhead, channel, host);
+	rc = write(proc->tofd, buf, len);
+
+	if (rc != len) {
+	  /* Failed ?  How come ?
+	     Close the 'tofd' pipe, and clear proc ptr. */
+	  pipes_shutdown_child(proc->tofd);
+	  proc->tofd = -1;
+	  proc->vertex->proc = NULL;
+	  proc->vertex       = NULL;
+	  proc->thread->proc = NULL;
+	  proc->thread       = NULL;
+	} else
+	  proc->fed = 1;
 }
 
-static void dotransport(files, n, vhead, channel, host)
-	char *files[];
-	int n;
+/*
+ * start_child() -- build argv[], and do other inits for fork()ing
+ *                  and execve()ing a new transport program for us.
+ */
+
+int
+start_child(vhead, chwp, howp)
 	struct vertex *vhead;
-	char *channel, *host;
-{
-	struct spblk *spl;
 	struct web *chwp, *howp;
+{
 	char	*av[30], *s, *os, *cp, *ocp, buf[MAXPATHLEN];
 	int	i;
 
-	if (n == 0)
-		return;
-	chwp = NULL;
-	if (channel != NULL) {
-	  spl = sp_lookup(symbol((u_char *)channel), spt_mesh[L_CHANNEL]);
-	  if (spl == NULL || (chwp = (struct web *)spl->data) == NULL)
-	    return;		/* yes, this can happen */
-	}
-	howp = NULL;
-	if (host != NULL) {
-	  spl = sp_lookup(symbol((u_char *)host), spt_mesh[L_HOST]);
-	  if (spl == NULL || (howp = (struct web *)spl->data) == NULL)
-	    return;		/* yes, this can happen */
-	}
+	if (verbose)
+	  printf("transport(vhead,chan=%s,host=%s)\n",
+		 chwp->name,howp->name);
+
 	if (vhead->ce.argv == NULL) {
-	  fprintf(stderr, "No command defined for %s/%s!\n", channel, host);
-	  return;
+	  fprintf(stderr, "No command defined for %s/%s!\n",
+		  chwp->name, howp->name);
+	  return 0;
 	}
 	/*
 	 * Replace the $host and $channel strings in the command line.
@@ -139,9 +122,9 @@ static void dotransport(files, n, vhead, channel, host)
 	os = buf;
 	for (i = 0; vhead->ce.argv[i] != NULL; ++i) {
 	  if (cistrcmp(vhead->ce.argv[i], replhost) == 0) {
-	    av[i] = host;
+	    av[i] = howp->name;
 	  } else if (cistrcmp(vhead->ce.argv[i], replchannel) == 0) {
-	    av[i] = channel;
+	    av[i] = chwp->name;
 	  } else if (strchr(vhead->ce.argv[i], '$') != NULL) {
 	    s = os;
 	    for (cp = vhead->ce.argv[i]; *cp != '\0'; ++cp) {
@@ -167,46 +150,24 @@ static void dotransport(files, n, vhead, channel, host)
 	    av[i] = vhead->ce.argv[i];
 	}
 	av[i] = NULL;
-	/*
-	 * randomize the file names if appropriate, this is a simpleminded
-	 * but probably effective way of avoiding a single message blocking
-	 * some queue.  XX: This should really be expanded to invoke a general
-	 * strategy routine/mechanism.
-	 */
-	if (n > 1) {
-	  char *temp;
-	  u_int ni;
 
-	  for (i = 0; i < n; ++i) {
-	    ni = ranny(n-1);
-	    temp = files[i];
-	    files[i] = files[ni];
-	    files[ni] = temp;
-	  }
-	}
 	/* fork off the appropriate command with the appropriate stdin */
 	if (verbose) {
 	  printf("$");
 	  for (i = 0; av[i] != NULL; ++i)
 	    printf(" %s", av[i]);
-	  printf(" <<\\EOF\n");
-	  for (i = 0; i < n; ++i)
-	    printf("%s\n", files[i]);
-	  printf("\\EOF\n");
+	  printf("\n");
 	}
-	runcommand(files, n, av, vhead, chwp, howp);
+	return runcommand(av, vhead, chwp, howp);
 }
 
-static void runcommand(filearr, nelem, argv, vhead, chwp, howp)
-	char	*filearr[];
-	int	nelem;
+static int runcommand(argv, vhead, chwp, howp)
 	char	*argv[];
 	struct vertex *vhead;
 	struct web *chwp, *howp;
 {
-	int	i, pid, position, to[2], from[2], uid, gid;
+	int	i, pid, to[2], from[2], uid, gid;
 	char	*cmd;
-	FILE	*fp;
 	static int pipesize = 0;
 
 
@@ -214,32 +175,17 @@ static void runcommand(filearr, nelem, argv, vhead, chwp, howp)
 	gid = vhead->ce.gid;
 	cmd = vhead->ce.command;
 
-	if (nelem <= 0)
-	  return;
-	if (epipe(to) < 0)
-	  return;
-	if (epipe(from) < 0) {
-	  close(to[0]);
-	  close(to[1]);
-	  return;
-	}
+	if (pipes_create(to,from) < 0) return 0;
 	if (pipesize == 0)
 	  pipesize = resources_query_pipesize(to[0]);
 
 	if (verbose)
 	  fprintf(stderr, "to %d/%d from %d/%d\n",
 		  to[0],to[1],from[0],from[1]);
+
 	if ((pid = fork()) == 0) {	/* child */
-	  int topfileno;
-	  if (to[0] != 0) {
-	    dup2(to[0],0);	/* child stdin */
-	    close(to[0]);
-	  }
-	  if (from[1] != 1) {
-	    dup2(from[1],1);	/* child stdout */
-	    close(from[1]);
-	  }
-	  /* dup2(1,2); */
+
+	  pipes_to_child_fds(to,from);
 
 	  /* keep current stderr for child stderr */
 	  /* close all other open filedescriptors */
@@ -261,38 +207,20 @@ static void runcommand(filearr, nelem, argv, vhead, chwp, howp)
 	  close(to[0]); close(to[1]);
 	  close(from[0]); close(from[1]);
 	  fprintf(stderr, "Fork failed!\n");
-	  return;
-	} else {		/* parent */
-	  close(to[0]);
-	  close(from[1]);
-	  ++numkids;
-	  if (chwp != NULL) chwp->kids += 1;
-	  if (howp != NULL) howp->kids += 1;
-	  /* save from[0] away as a descriptor to watch */
-	  stashprocess(pid, from[0], to[1], chwp, howp, vhead);
-	  /* write the list of files into to[1], and close it */
-	  if ((fp = fdopen(to[1], "w")) == NULL) {
-	    if (verbose)
-	      fprintf(stderr, "cannot open fd %d for w.\n",
-		      to[1]);
-	    close(to[1]);
-	    close(from[0]);
-	    return;
-	  }
-	  for (i = position = 0; i < nelem; ++i) {
-	    position += strlen(filearr[i]) + 1;
-	    if (position + 1 >= pipesize)
-	      break;
-	    fprintf(fp, "%s\n", filearr[i]);
-	  }
-	  if (verbose)
-	    fprintf(stderr, "wrote %d bytes to fd %d\n",
-		    position, fileno(fp));
-	  fputc('\n', fp);
-	  fclose(fp);	/* This will also close the to[1] fd */
+	  return 0;
 	}
+
+	/* parent */
+
+	pipes_close_parent(to,from);
+
+	/* save from[0] away as a descriptor to watch */
+	stashprocess(pid, from[0], to[1], chwp, howp, vhead);
+	/* We wait for the child to report "#hungry", then we feed it.. */
+	return 1;
 }
 
+
 static void stashprocess(pid, fromfd, tofd, chwp, howp, vhead)
 	int pid, fromfd, tofd;
 	struct web *chwp, *howp;
@@ -311,21 +239,27 @@ static void stashprocess(pid, fromfd, tofd, chwp, howp, vhead)
 	  i = scheduler_nofiles;
 	  cpids = (struct procinfo *)
 	    emalloc((unsigned)(i * sizeof (struct procinfo)));
-	  while (--i >= 0)
-	    cpids[i].pid = 0;
+	  memset(cpids, 0, sizeof(struct procinfo) * i);
 	}
-
 	proc = &cpids[fromfd];
-
+	proc->next   = NULL;
 	proc->pid    = pid;
 	proc->reaped = 0;
 	proc->ch     = chwp;
 	proc->ho     = howp;
-	proc->vertices = vhead;
+	proc->vertex = vhead;
+	proc->thread = vhead->thread;
+	proc->thread->proc = proc;
+	proc->thg    = vhead->thread->thgrp;
+	proc->thg->transporters += 1;
+	++numkids;
+	if (chwp != NULL) chwp->kids += 1;
+	if (howp != NULL) howp->kids += 1;
 	proc->tofd   = tofd;
 	vhead->proc  = proc;
 	proc->carryover = NULL;
-	proc->hungry = 1;
+	proc->hungry = 0;
+	proc->fed    = 0;
 #ifdef O_NONBLOCK /* POSIXy thing */
 	/* set up non-blocking I/O */
 	i = fcntl(fromfd, F_GETFL, 0);
@@ -343,116 +277,116 @@ static void stashprocess(pid, fromfd, tofd, chwp, howp, vhead)
 #endif	/* O_NONBLOCK  */
 }
 
-static void reclaim(fromfd)
-	int fromfd;
+static void reclaim(fromfd, tofd)
+	int fromfd, tofd;
 {
-	struct vertex *vp, *nvp;
-
+	struct vertex *vp, *nvp = NULL;
 	struct procinfo *proc = &cpids[fromfd];
 
+if (verbose)
+  fprintf(stderr,"reclaim(%d,%d) pid=%d, reaped=%d, chan=%s, host=%s\n",
+	  fromfd,tofd,proc->pid,proc->reaped,proc->ch->name,proc->ho->name);
+
 	proc->pid = 0;
 	proc->reaped = 0;
 	if (proc->carryover != NULL) {
 	  fprintf(stderr, "%s: HELP! Lost %d bytes: '%s'\n",
-		  progname, strlen(proc->carryover),
-		  proc->carryover);
+		  progname, strlen(proc->carryover), proc->carryover);
 	  free(proc->carryover);
 	  proc->carryover = NULL;
 	}
 	if (proc->ch != NULL) {
-	  --proc->ch->kids;
-	  if (proc->ch->fifohead)
-	    fifoschedule(proc->ch);
-	  else if (proc->ch->kids == 0 && proc->ch->link == NULL) {
+	  proc->ch->kids -= 1;
+	  if (proc->ch->kids == 0 && proc->ch->link == NULL) {
 	    unweb(L_CHANNEL, proc->ch);
 	    proc->ch = NULL;
 	  }
 	}
 	if (proc->ho != NULL) {
-	  --proc->ho->kids;
-	  if (proc->ho->fifohead)
-	    fifoschedule(proc->ho);
-	  else if (proc->ho->kids == 0 && proc->ho->link == NULL) {
+	  proc->ho->kids -= 1;
+	  if (proc->ho->kids == 0 && proc->ho->link == NULL) {
 	    unweb(L_HOST, proc->ho);
 	    proc->ho = NULL;
 	  }
 	}
+	if (tofd >= 0)
+	  pipes_shutdown_child(tofd);
 	close(fromfd);
 
 	/* Reschedule the vertices that are left
 	   (that were not reported on).		*/
 
-	nvp = proc->vertices;
-	proc->vertices = NULL;
-	for (vp = nvp; vp != NULL ; vp = nvp) {
-	  nvp = vp->nextitem;
-	  if (vp->ngroup == 0)
-	    abort();
-	  reschedule(vp, 0, -1);
+	/* ... but only if we were not in IDLE chain! */
+	if (proc->thread != NULL) {
+	  nvp = proc->thread->vertices;
+	  proc->thg->transporters -= 1;
+	  --numkids;
+	  proc->vertex = NULL;
+	  proc->thread = NULL;
+	  proc->thg    = NULL;
+	  /* XX: thread reschedule ?? */
+	  for (vp = nvp; vp != NULL ; vp = nvp) {
+	    nvp = vp->nextitem;
+	    if (vp->ngroup == 0)
+	      abort();
+	    reschedule(vp, 0, -1);
+	  }
+	} else {
+	  /* Maybe we were in idle chain! */
+	  if (proc->thg != NULL /* And not killed by  idle_cleanup() */) {
+	    struct procinfo *p, **pp;
+	    p  =  proc->thg->idleproc;
+	    pp = &proc->thg->idleproc;
+
+	    while (p && p != proc) {
+		/* Move to the next possible idle process */
+		pp = &p->next;
+		p = p->next;
+	    }
+	    if (p == proc) {
+	      /* Remove this entry from the chains */
+	      *pp = p->next;
+	      p = p->next;
+	      proc->thg->idlecnt      -= 1;
+	      --idleprocs;
+	      proc->thg->transporters -= 1;
+	      --numkids;
+	      /* It may go down to zero.. */
+	      if (proc->thg->transporters == 0 && proc->thg->threads == 0)
+		delete_threadgroup(proc->thg);
+	    } else {
+	      /* It is not in idle chain, it has died somehow else.. */
+	      if (proc->vertex != NULL)
+		proc->vertex->proc = NULL;
+	      proc->vertex = NULL;
+	      if (proc->thread != NULL)
+		proc->thread->proc = NULL;
+	      proc->thread = NULL;
+	      proc->thg->transporters -= 1;
+	      --numkids;
+	    }
+	  } else {
+	    /* We are killed by the idle_cleanup() !   */
+	    /* proc->thg == NULL, proc->thread == NULL */
+	    /* idle_cleanup() did all decrementing..   */
+	  }
 	}
 }
 
 static void waitandclose(fd)
 	int	fd;
 {
-#ifndef USE_SIGREAPER
-	int	n, pid;
-
-#if 0
-	/* This is a quick and dirty hack (for A/UX I think). BEWARE */
-	if (cpids[fd].pid <= 0) {
-	  fprintf(stderr,"fd %d: pid is %d ??\n",fd,cpids[fd].pid);
-	  return;
-	}
-	n = waitpid(cpids[fd].pid, (int *)NULL, 0);
-	if(n != cpids[fd].pid && (n != -1 || errno != 10)) {
-	  fprintf(stderr,"fd %d, pid %d: waitpid says %d, errno %d\n",
-		  fd,cpids[fd].pid, n, errno);
-	  return;
-	}
-	--numkids;
-	reclaim(fd);
-#else
-	if (cpids[fd].pid > 0) {
-#ifdef	USE_UNIONWAIT
-	  while ((pid = wait((union wait *)0)) > 0)
-#else	/* !USE_UNIONWAIT */
-	  while ((pid = wait((int *)0)) > 0)
-#endif	/* USE_UNIONWAIT */
-	    {
-	      --numkids;
-	      if (verbose)
-		printf("reaped %d\n", pid);
-	      if (pid == cpids[fd].pid) {
-		reclaim(fd);
-		break;
-	      } else {
-		for (n =; n < scheduler_nofiles; ++n) {
-		  if (cpids[n].pid == pid) {
-		    cpids[n].pid = -pid;
-		    ++numkids;
-		    break;
-		  }
-		}
-	      }
-	    }
-	} else if (cpids[fd].pid < 0) {
-	  --numkids;
-	  reclaim(fd);
-	}
-#endif
-#else	/* USE_SIGREAPER */
 	/* This is called when
 	   - fd return 0 (EOF)
 	   - fd returns -1, and errno != {EAGAIN|EWOULDBLOCK|EINTR}
 	 */
-	--numkids;
-	reclaim(fd);
-#endif
+	reclaim(fd, cpids[fd].tofd);
+
 }
 
 #ifdef	USE_SELECT
 
+
 #if	defined(BSD4_3) || defined(sun)
 #include <sys/file.h>
 #endif
@@ -513,6 +447,8 @@ typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
 #endif
 #endif
 
+int in_select = 0;
+
 int
 mux(timeout)
 time_t timeout;
@@ -523,13 +459,14 @@ time_t timeout;
 #else
 	fd_set	mask;
 #endif
-	struct sockaddr_in raddr;
-	int	raddrlen;
 	struct timeval tv;
 
+	if (in_select) fprintf(stderr,"**** recursed into mux()! ***\n");
+	in_select = 1;
+
 	time(&now);
 
-	tv.tv_sec = timeout - now;
+	tv.tv_sec = timeout - now; /* Timeout in seconds */
 	if (timeout < now)
 	  tv.tv_sec = 0;
 	tv.tv_usec = 0;
@@ -553,16 +490,34 @@ time_t timeout;
 	/*fprintf(stderr, "about to select on %x [%d]\n",
 			  mask.fds_bits[0], maxf); */
 	if ((n = select(maxf, &mask, NULL, NULL, &tv)) < 0) {
+	  int err = errno;
 	  /* fprintf(stderr, "got an interrupt (%d)\n", errno); */
-	  if (errno == EINTR)
+	  in_select = 0;
+	  if (errno == EINTR || errno == EAGAIN)
 	    return 0;
+	  if (errno == EINVAL || errno == EBADF) {
+	    fprintf(stderr, "** select() returned errno=%d\n", err);
+	    for (i = 0; i < maxf; ++i)
+	      if (_FD_ISSET(i,mask) &&
+		  fcntl(i,F_GETFL,0) < 0)
+		fprintf(stderr,"** Invalid fd on a select() mask: %d\n",i);
+	    fflush(stderr);
+	    abort();
+	  }
+	  perror("select() returned unknown error ");
+	  fflush(stderr);
+	  abort();
 	} else if (n == 0) {
 	  /* fprintf(stderr, "abnormal 0 return from select!\n"); */
 	  /* -- just a timeout -- fast or long */
+	  in_select = 0;
 	  return 1;
 	} else {
 	  /*fprintf(stderr, "got %d ready (%x)\n", n, mask.fds_bits[0]);*/
 	  if (querysocket >= 0 && _FD_ISSET(querysocket, mask)) {
+	    struct sockaddr_in raddr;
+	    int	raddrlen;
+
 	    --n;
 	    _FD_CLR(querysocket, mask);
 	    raddrlen = sizeof raddr;
@@ -577,8 +532,8 @@ time_t timeout;
 		execl("/bin/false",NULL);
 		_exit(0);
 	      }
-	      if (rc > 0)
-		++numkids;
+	      /* if (rc > 0)
+		 ++numkids; */
 	      close(i);
 	    }
 	  }
@@ -594,9 +549,48 @@ time_t timeout;
 	    }
 	}
 	/* fprintf(stderr, "return from mux\n"); */
+	in_select = 0;
 	return 0;
 }
 
+void
+queryipccheck()
+{
+	int	n;
+#ifdef _AIX
+	struct { fd_set fdsmask; fd_set msgsmask; } mask;
+#else
+	fd_set	mask;
+#endif
+	struct timeval tv;
+
+	tv.tv_sec = 0;
+	tv.tv_usec = 0;
+
+	_FD_ZERO(mask);
+	if (querysocket >= 0) {
+	  _FD_SET(querysocket, mask);
+
+	  n = select(querysocket+1, &mask, NULL, NULL, &tv);
+	  if (n > 0 &&
+	      _FD_ISSET(querysocket, mask)) {
+	    struct sockaddr_in raddr;
+	    int raddrlen = sizeof(raddr);
+
+	    n = accept(querysocket, (struct sockaddr *)&raddr, &raddrlen);
+	    if (n >= 0) {
+	      int pid = fork();
+	      if (pid == 0) {
+		qprint(n);
+		execl("/bin/false",NULL);
+		_exit(0);
+	      }
+	      close(n);
+	    }
+	  }
+	}
+}
+
 void
 queryipcinit()
 {
@@ -625,13 +619,13 @@ queryipcinit()
 	setsockopt(querysocket, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
 
 	if (bind(querysocket, (struct sockaddr *)&sad, sizeof sad) < 0) {
-	  perror("bind");
+	  perror("bind:mailq socket");
 	  close(querysocket);
 	  querysocket = -1;
 	  return;
 	}
 	if (listen(querysocket, 5) < 0) {
-	  perror("listen");
+	  perror("listen:mailq socket");
 	  close(querysocket);
 	  querysocket = -1;
 	  return;
@@ -664,6 +658,11 @@ time_t timeout;
 	return 1;
 }
 
+void queryipccheck()
+{
+	/* NOTHING AT ALL -- No select(), no querysocket.. */
+}
+
 void
 queryipcinit()
 {
@@ -675,26 +674,28 @@ static void readfrom(fd)
 {
 	int	n, dontbreak, bufsize = 2048;
 	char	*cp, *pcp, *eobuf, *buf;
+	struct procinfo *proc = &cpids[fd];
 
 	dontbreak = 0;
 	cp = pcp = NULL;
 
 	buf = (char *)emalloc(bufsize);
 
-	if (cpids[fd].carryover != NULL) {
-	  int carrylen = strlen(cpids[fd].carryover);
+	if (proc->carryover != NULL) {
+	  int carrylen = strlen(proc->carryover);
 	  if (carrylen > bufsize) {
 	    while (carrylen > bufsize)
 	      bufsize += 1024;
 	    buf = erealloc(buf,bufsize);
 	  }
-	  strcpy(buf, cpids[fd].carryover);
+	  strcpy(buf, proc->carryover);
 	  cp = buf+strlen(buf);
 	  pcp = buf;
-	  free(cpids[fd].carryover);
-	  cpids[fd].carryover = NULL;
+	  free(proc->carryover);
+	  proc->carryover = NULL;
 	  dontbreak = 1;
 	}
+
 	/* Note that if we get an alarm() call, the read will return -1, TG */
 	errno = 0;
 	while ((n = read(fd, dontbreak ? cp : buf,
@@ -707,8 +708,8 @@ static void readfrom(fd)
 	    if (*cp == '\n') {
 	      *cp = '\0';
 	      if (verbose)
-		fprintf(stderr, "%d processed: %s\n",
-			fd, pcp);
+		fprintf(stderr, "%d fd=%d processed: %s\n",
+			proc->pid,fd, pcp);
 	      update(fd,pcp);
 	      *cp = '_';
 	      pcp = cp + 1;
@@ -716,6 +717,7 @@ static void readfrom(fd)
 	    } else
 	      dontbreak = 1;
 	  }
+
 	  if (dontbreak && cp == buf + bufsize) {
 	    if (pcp == buf) {
 	      /* XX:
@@ -739,12 +741,18 @@ static void readfrom(fd)
 	    break;
 #endif				/* !USE_BSDNODELAY */
 	}
+
 	if (verbose) {
-	  fprintf(stderr, "read from %d returns %d, errno=%d\n", fd, n, errno);
+	  if (!(errno == EAGAIN || errno == EWOULDBLOCK))
+	    fprintf(stderr,
+		    "read from %d returns %d, errno=%d\n", fd, n, errno);
 	}
 	if (n == 0 || (n < 0 && errno != EWOULDBLOCK && errno != EAGAIN &&
 		       errno != EINTR)) {
 	  /*printf("about to call waitandclose(), n=%d, errno=%d\n",n,errno);*/
+	  if (proc->tofd >= 0)
+	    pipes_shutdown_child(proc->tofd);
+	  proc->tofd = -1;
 	  waitandclose(fd);
 	}
 	/* fprintf(stderr, "n = %d, errno = %d\n", n, errno); */
@@ -756,14 +764,14 @@ static void readfrom(fd)
 	 * or else some report may get lost somewhere.
 	 */
 	if (dontbreak) {
-	  if (cpids[fd].pid != 0) {
-	    cpids[fd].carryover = emalloc(cp-pcp+1);
-	    memcpy(cpids[fd].carryover, pcp, cp-pcp);
-	    cpids[fd].carryover[cp-pcp] = '\0';
+	  if (proc->pid != 0) {
+	    proc->carryover = emalloc(cp-pcp+1);
+	    memcpy(proc->carryover, pcp, cp-pcp);
+	    proc->carryover[cp-pcp] = '\0';
 	  } else
 	    fprintf(stderr,
-		    "HELP! Lost %d bytes (n=%d/%d, off=%d): '%s'\n",
-		    cp - pcp, n, errno, pcp-buf, pcp);
+		    "HELP! Lost %ld bytes (n=%d/%d, off=%ld): '%s'\n",
+		    (long)(cp - pcp), n, errno, (long)(pcp-buf), pcp);
 	}
 	free(buf);
 }
@@ -848,51 +856,56 @@ int signum;
 #else
 	int statloc;
 #endif
+
+	for (;;) {
+
 #ifdef USE_UNIONWAIT
-	while ((pid = wait3(&statloc,WNOHANG,NULL)) > 0) ;
+	  pid = wait3(&statloc,WNOHANG,NULL);
 #else
 #ifdef	USE_WAITPID
-	while ((pid = waitpid(-1,&statloc,WNOHANG)) > 0) ;
+	  pid = waitpid(-1,&statloc,WNOHANG);
 #else
-	pid = wait(&statloc);
+	  pid = wait(&statloc);
 #endif
 #endif
-
-
-#if 0	/* DO NOT do anything excessive -- we don't care WHO died.. */
+	  if (pid <= 0) return;
 
 #ifdef	WIFEXITED
-	if (WIFEXITED(statloc)) ok = 1;
+	  if (WIFEXITED(statloc)) ok = 1;
 #else
-      ERROR:ERROR:ERROR
+      ERROR_ERROR_ERROR:
 #endif
 #ifdef WIFSIGNALED
 	if (WIFSIGNALED(statloc)) ok = 1;
 #else
-      ERROR:ERROR:ERROR
+      ERROR_ERROR_ERROR:
 #endif
 
-	if (verbose)
-	  fprintf(stderr,"sig_chld() pid=%d, ok=%d\n",pid,ok);
+	  if (verbose)
+	    fprintf(stderr,"sig_chld() pid=%d, ok=%d, stat=0x%x\n",
+		    pid,ok,statloc);
+	  
+	  if (ok && cpids != NULL) {
+	    /* Only EXIT and SIGxx DEATHS accepted */
 
-	if (ok && cpids != NULL) {
-	  /* Only EXIT and SIGxx DEATHS accepted */
-
-	  for (i = scheduler_nofiles-1; i >= 0; --i) {
-	    if (cpids[i].pid == pid) {
-	      cpids[i].pid = -pid;	/* Mark it as reaped.. */
-	      ok = 0;
-	      break;
-	    }
-	    if (cpids[i].pid == -pid) {
-printf(" .. already reaped ??\n");
-	      cpids[i].pid = -pid;	/* Mark it as reaped.. */
-	      ok = 0;
-	      break;
+	    for (i = scheduler_nofiles-1; i >= 0; --i) {
+	      if (cpids[i].pid == pid) {
+		cpids[i].pid = -pid; /* Mark it as reaped.. */
+		cpids[i].reaped = 1;
+		ok = 0;
+		break;
+	      }
+	      if (cpids[i].pid == -pid) {
+		printf(" .. already reaped ??\n");
+		cpids[i].pid = -pid; /* Mark it as reaped.. */
+		cpids[i].reaped = 1;
+		ok = 0;
+		break;
+	      }
 	    }
 	  }
 	}
-#endif
+
 	/* re-instantiate the signal handler.. */
 #ifdef SIGCLD
 	SIGNAL_HANDLE(SIGCLD,  sig_chld);
