@@ -16,7 +16,7 @@
 # define FSTATFS fstatvfs /* Existence of this means (usually) SYSV code */
 # define STATFSTYPE struct statvfs
 #else	/* !USE_STATVFS */
-#ifdef	SYSV
+#if defined(SYSV) || defined(RISC6000)
 # include <sys/statfs.h>
 # define FSTATFS fstatfs /* Existence of this means (usually) SYSV code */
 # define STATFSTYPE struct statfs
@@ -60,7 +60,7 @@ int fd;
 	  availspace = statbuf.f_bavail * statbuf.f_bsize;
 	}
 #else
-#if defined(ULTRIX)
+#if defined(ULTRIX) || defined(RISC6000)
 	if ((rc = FSTATFS(fd, &statbuf)) == 0) {
 	  availspace = statbuf.f_bavail * statbuf.f_bsize;
 	}
