@@ -166,9 +166,9 @@ char	myhostname[MAXHOSTNAMELEN+1];
 char	rhostname[MAXHOSTNAMELEN+1];
 char	ihostaddr[3*4+1*3+2+1 /* == 18 for '[255.255.255.255]' */ +10/*slack*/];
 int	rport;
-int	skeptical = 1;
-int	verbose = 0;
-int	daemon_flg	= 1;
+int	skeptical  = 1;
+int	verbose    = 0;
+int	daemon_flg = 1;
 int	pid, routerpid;
 FILE	*logfp = NULL;
 int	D_alloc = 0;
@@ -785,7 +785,7 @@ char *cp;
 	 * tokenizer -- RFC821SCN.C, lets use it :)
 	 * We need it for proper handling of ESMTP anyway
 	 */
-	if (partridge(helobuf, cfinfo)) {
+	if (skeptical && partridge(helobuf, cfinfo)) {
 	  rfc821_error_ptr += (cp - helobuf);
 	  type821err(501, buf,
 		     "Invalid `%s' parameter!  Err: %s",
@@ -801,7 +801,7 @@ char *cp;
 	  return;
 	}
 	/* Check helobuf corresponds to the reverse address */
-	if (ihostaddr[0] != '\0'
+	if (skeptical && ihostaddr[0] != '\0'
 	    && rhostname[0] != '\0' && rhostname[0] != '['
 	    && cistrcmp(helobuf, rhostname) != 0) {
 	  type(-250, " %s That hostname is inconsistent with",
@@ -1488,6 +1488,10 @@ smtpserver(insecure)
 		break;
 	    case Verify:
 	    case Verify2:
+		if (state == Hello) {
+		  type (503, "Waiting for HELO/EHLO command");
+		  break;
+		}
 		if ((cfi = STYLE(cfinfo, 'v'))) {
 		  while (*cp == ' ' || *cp == '\t') ++cp;
 		  if (*cp == '<')
@@ -1511,6 +1515,10 @@ smtpserver(insecure)
 		  type(252, (char *)NULL);
 		break;
 	    case Expand:
+		if (state == Hello) {
+		  type (503, "Waiting for HELO/EHLO command");
+		  break;
+		}
 		if ((cfi = STYLE(cfinfo, 'e'))) {
 		  while (*cp == ' ' || *cp == '\t') ++cp;
 		  if (*cp == '<')
@@ -1566,8 +1574,8 @@ smtpserver(insecure)
 	reporterr(ftell(stdin), "session terminated");
 }
 
-#define	TYPE_(m)	type(-241, "%s", m);
-#define	TYPE(m)		type(241, "%s", m);
+#define	TYPE_(m)	type(-214, "%s", m);
+#define	TYPE(m)		type(214, "%s", m);
 
 /*
  * parse the query string and print an appropriate help message.
