@@ -160,13 +160,13 @@ process(dp)
 	   * instead just drop it on the floor.
 	   */
 	  for (rp = dp->recipients; rp != NULL; rp = rp->next)
-	    diagnostic(rp, EX_OK, "error bounce dropped");
+	    diagnostic(rp, EX_OK, 0, "error bounce dropped");
 	  return;
 	}
 
 	if ((mfp = mail_open(MSG_RFC822)) == NULL) {
 	  for (rp = dp->recipients; rp != NULL; rp = rp->next)
-	    diagnostic(rp, EX_TEMPFAIL, "mail_open failure");
+	    diagnostic(rp, EX_TEMPFAIL, 0, "mail_open failure");
 	  warning("Cannot open mail file!");
 	  return;
 	}
@@ -232,7 +232,7 @@ process(dp)
 	if (reportcnt == 0){
 	  /* No, throw it away and ack success. */
 	  mail_abort(mfp);
-	  diagnostic(rp, EX_OK, NULL);
+	  diagnostic(rp, EX_OK, 0, NULL);
 	  return;
 	}
 
@@ -294,5 +294,5 @@ process(dp)
 	else
 	  n = EX_OK;
 	for (rp = dp->recipients; rp != NULL; rp = rp->next)
-	  diagnostic(rp, n, (char *)NULL);
+	  diagnostic(rp, n, 0, (char *)NULL);
 }
