@@ -97,10 +97,6 @@ typedef char msgdata;
 #include <sys/wait.h>
 #endif	/* USE_UNIONWAIT */
 
-#ifndef	IPPORT_SMTP
-#define	IPPORT_SMTP	25
-#endif	/* IPPORT_SMTP */
-
 #ifndef	L_SET
 #define	L_SET	0
 #endif	/* L_SET */
@@ -132,6 +128,8 @@ extern time_t time();
 extern char **environ;
 
 extern char *strchr(), *strrchr();
+extern struct hostent *gethostbyname(), *gethostbyaddr();
+extern int ranny();
 
 #ifdef	lint
 #undef	putc
@@ -460,7 +458,6 @@ hold_ns(s)
 {
 	struct qtypes *qtp;
 	char *cp, host[BUFSIZ];
-	extern int getrrtype();
 
 	if ((cp = strrchr(s, '/')) == NULL)
 	  return 1;	/* human error, lets be nice */
@@ -480,7 +477,7 @@ hold_ns(s)
 	  return 1;		/* inconsistency with search_res.c, yell! */
 	}
 	strcpy(host, s);
-	switch (getrrtype(host, sizeof host, qtp->value)) {
+	switch (getrrtype(host, sizeof host, qtp->value, 2)) {
 	case 0:
 		return 1;	/* negative reply */
 	case 1:
@@ -513,7 +510,6 @@ hold_ns(s)
 	char *cp, host[BUFSIZ];
 	struct hostent *hp;
 	struct in_addr ia;
-	extern struct hostent *gethostbyname(), *gethostbyaddr();
 
 	if ((cp = strrchr(s, '/')) == NULL)
 	  return 1;	/* human error, lets be nice */
@@ -569,8 +565,6 @@ int
 hold_io(s)
 	char *s;
 {
-	extern int ranny();
-	
 	return ranny(9) == 0;	/* 10% of the time */
 }
 
@@ -721,134 +715,3 @@ char *user;
 	    S_ISDIR(st.st_mode)) return 1;
 	return 0;
 }
-
-#ifdef	BIND
-
-typedef union {
-	HEADER qb1;
-	char qb2[PACKETSZ];
-} querybuf;
-
-extern int h_errno;
-
-int
-getrrtype(host, hbsize, rrtype)
-	char *host;
-	int hbsize;
-	u_short rrtype;
-{
-
-	HEADER *hp;
-	msgdata *eom, *cp;
-	querybuf buf, answer;
-	int qlen, n, ancount, qdcount, ok;
-	u_short type;
-	msgdata nbuf[BUFSIZ];
-	int first;
-	extern int res_mkquery(), res_send(), dn_expand(), dn_skipname();
-
-	qlen = res_mkquery(QUERY, host, C_IN, rrtype, (char *)NULL, 0, NULL,
-			   (char *)&buf, sizeof(buf));
-	if (qlen < 0) {
-	  fprintf(stderr, "res_mkquery failed\n");
-	  h_errno = NO_RECOVERY;
-	  strcpy(errormsg, "no recovery");
-	  return -2;
-	}
-	n = res_send((char *)&buf, qlen, (char *)&answer, sizeof(answer));
-	if (n < 0) {
-	  h_errno = TRY_AGAIN;
-	  strcpy(errormsg, "try again");
-	  return -1;
-	}
-	eom = (msgdata *)&answer + n;
-	/*
-	 * find first satisfactory answer
-	 */
-	hp = (HEADER *) &answer;
-	ancount = ntohs(hp->ancount);
-	qdcount = ntohs(hp->qdcount);
-	h_errno = 0;
-	/*
-	 * We don't care about errors here, only if we got an answer
-	 */
-	if (ancount == 0) {
-	  if (rrtype == T_CNAME && hp->rcode == NOERROR) {
-	    if (qdcount > 0 && strchr(host, '.') == NULL) {
-	      cp = (msgdata *)&answer + sizeof(HEADER);
-	      if (dn_expand((msgdata *)&answer,
-			    eom, cp, host, hbsize) >= 0) {
-		if (host[0] == '\0') {
-		  host[0] = '.';
-		  host[1] = '\0';
-		}
-	      }
-	    }
-	    return 1;
-	  }
-	  return (hp->rcode == NOERROR || hp->rcode == NXDOMAIN) ? 0 : -3;
-	}
-	cp = (msgdata *)&answer + sizeof(HEADER);
-	for (; qdcount > 0; --qdcount)
-#if	defined(BIND_VER) && (BIND_VER >= 473)
-		cp += dn_skipname(cp, eom) + QFIXEDSZ;
-#else	/* !defined(BIND_VER) || (BIND_VER < 473) */
-		cp += dn_skip(cp) + QFIXEDSZ;
-#endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
-	first = 1;
-	ok = rrtype != T_WKS;
-	while (--ancount >= 0 && cp < eom) {
-	  if ((n = dn_expand((msgdata *)&answer, eom, cp, nbuf,
-			     sizeof(nbuf))) < 0)
-	    break;
-	  if (first) {
-	    strncpy(host, (char *)nbuf, hbsize);
-	    host[hbsize - 1] = '\0';
-	    first = 0;
-	  }
-	  cp += n;
-	  type = _getshort(cp);
-	  cp += sizeof(u_short); /* type */
-	  cp += sizeof(u_short); /* class */
-	  cp += sizeof(u_long);	/* ttl */
-	  n = _getshort(cp);
-	  cp += sizeof(u_short); /* dlen */
-	  if (type != rrtype) {
-	    cp += n;
-	    continue;
-	  }
-	  /*
-	   * Assume that only one rrtype will be found.  More
-	   * than one is undefined.
-	   */
-	  if (type == T_WKS) {
-	    msgdata *nextcp = cp + n;
-	    /* If we have seen a WKS, it had better have SMTP,
-	     * however in absence of a WKS, assume SMTP.
-	     */
-	    if (n < sizeof (u_long) + 1) {
-	      cp = nextcp;	/* ? */
-	      continue;
-	    }
-	    ok = 0;
-	    cp += sizeof(u_long); /* skip IP address */
-	    if (*cp++ == IPPROTO_TCP) {	/* check protocol */
-	      if (cp + (IPPORT_SMTP/8) < nextcp
-		  && (*(cp+(IPPORT_SMTP/8))
-		      & (0x80>>IPPORT_SMTP%8)))
-		return 1;
-	    }
-	    cp = nextcp;
-	    continue;
-	  } else {
-	    if ((n = dn_expand((msgdata *)&answer, eom, cp, nbuf,
-			       sizeof(nbuf))) < 0)
-	      break;
-	    strncpy(host, (char *)nbuf, hbsize);
-	    host[hbsize - 1] = '\0';
-	  }
-	  return ok;
-	}
-	return 0;
-}
-#endif	/* BIND */
