@@ -6,6 +6,8 @@
 #include <stdio.h>
 #include "hostenv.h"
 
+extern char *strchr();
+
 int
 emptyline(line, size)
 	char *line;
@@ -16,7 +18,7 @@ emptyline(line, size)
 	if (line[0] == '\0' || line[0] == '\n')
 		return 1;
 	line[size - 1] = '\0';
-	if ((s = (char*)strchr(line, '\n')) == NULL)
+	if ((s = strchr(line, '\n')) == NULL)
 		return 1;
 	*s = '\0';
 	return 0;
