@@ -929,7 +929,7 @@ if (verboselog) fprintf(verboselog,"8-bit header: '%s'\n",hdr);
 	      strcpy(q,"=?UNKNOWN-8BIT?Q?");
 	      q += 17; /* 17 chars above */
 
-	      for ( ; *s && (*s != ' ' && *s != '\t'); ++s) {
+	      for ( ; *s && (*s != ' ' && *s != '\t' && *s != ')'); ++s) {
 		if (*s < ' '  || *s > 126  ||
 		    *s == '=' || *s == '?' || *s == '_') {
 		  sprintf(q,"=%02X",*s);
