@@ -28,9 +28,9 @@ writeheaders(rp,fp,newline,convertmode,maxwidth)
 
 	/* If asked for, and they are available,
 	   use converted headers */
-	if (convertmode != _CONVERT_NONE)
-	  if (*(rp->newmsgheadercvt) != NULL)
-	    msgheaders = *(rp->newmsgheadercvt);
+	/* if (convertmode != _CONVERT_NONE) */
+	if (*(rp->newmsgheadercvt) != NULL)
+	  msgheaders = *(rp->newmsgheadercvt);
 
 	if (!msgheaders) return -1;
 
