@@ -271,6 +271,12 @@ struct conshell *envarlist = NULL;
 extern int stickymem;	/* for strsave() */
 int	D_alloc = 0;
 
+static void sig_alrm(sig)
+int sig;
+{
+	/* Sigh, actually dummy routine.. */
+}
+
 
 #ifndef	MAXPATHLEN
 #define	MAXPATHLEN 1024
@@ -284,12 +290,13 @@ main(argc, argv)
 	char file[MAXPATHLEN+128];
 	char *s;
 	int c, errflg, fd;
-	struct ctldesc *dp;
 	char *host = NULL;	/* .. and what is my host ? */
 	int matchhost = 0;
 #if	defined(BIFF) || defined(RBIFF)
 	struct biffer *nbp;
 #endif
+	struct ctldesc *dp;
+
 	SIGNAL_TYPE (*oldsig)();
 
 	SIGNAL_HANDLESAVE(SIGINT, SIG_IGN, oldsig);
@@ -301,6 +308,8 @@ main(argc, argv)
 	SIGNAL_HANDLESAVE(SIGHUP, SIG_IGN, oldsig);
 	if (oldsig != SIG_IGN)
 	  SIGNAL_HANDLE(SIGHUP, wantout);
+	SIGNAL_HANDLE(SIGALRM, sig_alrm); /* Actually ignored, but
+					     fcntl() will break ? */
 
 	SIGNAL_IGNORE(SIGPIPE);
 
@@ -402,6 +411,9 @@ main(argc, argv)
 	    *s = 0;
 	  }
 
+	  setreuid(0,0); /* We begin as roots..  process() may change us */
+	  currenteuid = 0;
+
 	  dp = ctlopen(file, channel, host, &getout, NULL, NULL);
 	  if (dp != NULL) {
 	    if (verboselog) {
@@ -416,33 +428,43 @@ main(argc, argv)
 	    process(dp);
 	    ctlclose(dp);
 	  }
-	}
 
 #if	defined(BIFF) || defined(RBIFF)
-	/* now that the delivery phase is over, hoot'n wave */
-	eobiffs = NULL;
-	for (nbp = biffs ; nbp != NULL; nbp = nbp->next) {
-	  if (dobiff && nbp->offset >= 0)
-	    biff("localhost", nbp->user, nbp->offset);
+	  /* now that the delivery phase is over, hoot'n wave */
+	  for (nbp = biffs ; nbp != NULL; nbp = nbp->next) {
+	    if (dobiff && nbp->offset >= 0)
+	      biff("localhost", nbp->user, nbp->offset);
 #ifdef	RBIFF
-	  if (nbp->wantrbiff != 0) {
-	    if (spt_users == NULL)
-	      spt_users = sp_init();
-	    sp_install(symbol(nbp->user),
-		       (char *)NULL, 0, spt_users);
-	  }
-#endif /* RBIFF */
-	}
-#ifdef	RBIFF
-	if (spt_users != NULL) {
-	  if (readrwho())
-	    for (nbp = biffs ; nbp != NULL; nbp = nbp->next) {
-	      if (nbp->offset >= 0 && nbp->wantrbiff)
-		rbiff(nbp);
+	    if (nbp->wantrbiff != 0) {
+	      if (spt_users == NULL)
+		spt_users = sp_init();
+	      sp_install(symbol(nbp->user),
+			 (char *)NULL, 0, spt_users);
 	    }
-	}
+#endif /* RBIFF */
+	  }
+#ifdef	RBIFF
+	  if (spt_users != NULL) {
+	    if (readrwho())
+	      for (nbp = biffs ; nbp != NULL; nbp = nbp->next) {
+		if (nbp->offset >= 0 && nbp->wantrbiff)
+		  rbiff(nbp);
+	      }
+	  }
 #endif	/* RBIFF */
+	  while ((nbp = biffs) != NULL) {
+	    biffs = nbp->next;
+	    free(nbp->user);
+	    free(nbp);
+	  }
+	  {
+	    extern void free();
+	    sp_scan(free,NULL,spt_users);
+	    free(spt_users);
+	    spt_users = NULL;
+	  }
 #endif	/* BIFF || RBIFF */
+	}
 	exit(EX_OK);
 	/* NOTREACHED */
 	return 0;
@@ -1784,7 +1806,7 @@ readrwho()
 	
 	if (spt_users == NULL)
 	  return;
-	now = time(0);
+	now = time(NULL);
 	if (chdir(RWHODIR) || (dirp = opendir(".")) == NULL ) {
 	  /* BE SILENT -- failing RBIFF is no fault! */
 	  return 0;
