@@ -375,14 +375,14 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	/* now we can fork off and run the command... */
 	if (pipe(in) < 0) {
 	  for (rp = startrp; rp != endrp; rp = rp->next)
-	    diagnostic(rp, EX_OSERR,
+	    diagnostic(rp, EX_OSERR, 0,
 		       "cannot create pipe from \"%s\"",
 		       mp->command);
 	  return;
 	}
 	if (pipe(out) < 0) {
 	  for (rp = startrp; rp != endrp; rp = rp->next)
-	    diagnostic(rp, EX_OSERR,
+	    diagnostic(rp, EX_OSERR, 0,
 		       "cannot create pipe to \"%s\"",
 		       mp->command);
 	  return;
@@ -405,7 +405,7 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	  _exit(254);
 	} else if (pid < 0) {	/* couldn't fork, complain */
 	  for (rp = startrp; rp != endrp; rp = rp->next)
-	    diagnostic(rp, EX_OSERR, "cannot fork");
+	    diagnostic(rp, EX_OSERR, 0, "cannot fork");
 	  return;
 	}
 	close(out[0]);
@@ -508,7 +508,7 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	/* append message body itself */
 	if ((i = appendlet(dp, mp, tafp, verboselog, convertmode)) != EX_OK) {
 	  for (rp = startrp; rp != endrp; rp = rp->next)
-	    diagnostic(rp, i, "write error");
+	    diagnostic(rp, i, 0, "write error");
 	  /* just to make sure nothing will get delivered */
 	  kill(pid, SIGTERM);
 	  sleep(1);
@@ -572,7 +572,7 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	  if (verboselog)
 	    fprintf(verboselog,"Diagnostic: %s\n",cp);
 	  for (rp = startrp; rp != endrp; rp = rp->next) {
-	    diagnostic(rp, i, "%s", buf);
+	    diagnostic(rp, i, 0, "%s", buf);
 	  }
 	  /* XX: still need to deal with MO_STRIPQUOTES */
 	}
