@@ -170,6 +170,7 @@ char myhostname[MAXHOSTNAMELEN+1];
 char remotehost[MAXHOSTNAMELEN+1];
 char remotemsg[2*BUFSIZ];
 char ipaddress[30];
+char errormsg[BUFSIZ];
 char *progname;
 char *cmdline, *eocmdline, *logfile, *msgfile;
 char *myaddrs = NULL;
@@ -265,13 +266,11 @@ extern int rightmx __((char*,char*));
 extern int h_errno;
 extern int res_mkquery(), res_send(), dn_skipname(), dn_expand();
 # ifdef RFC974
-extern int getrr __((char*,int,u_short,int));
 struct mxdata {
 	msgdata	*host;
 	u_short	pref;
 };
 extern int getmxrr __((char*,struct mxdata*,int,int,int*));
-extern int getrrtype __((char*,int,u_short,int));
 # endif /* RFC974 */
 #endif	/* BIND */
 extern int matchroutermxes __((char*,struct address*));
@@ -472,7 +471,7 @@ main(argc, argv)
 	      if (smtp_isopen) {
 		smtpstatus = smtpwrite(smtpfp, verboselog, 0, "QUIT");
 		fclose(smtpfp[0]); smtpfp[0] = NULL;
-		fclose(smtpfp[0]); smtpfp[0] = NULL;
+		fclose(smtpfp[1]); smtpfp[1] = NULL;
 		if (logfp)
 		  fprintf(logfp, "%d#\t(closed SMTP channel - new host)\n", pid);
 		smtp_isopen = 0;
@@ -504,10 +503,6 @@ main(argc, argv)
 
 	  headers_contain_8bit = 0;
 
-	  if (verboselog) {
-	    fclose(verboselog);
-	    verboselog = NULL;
-	  }
 	  if (dp->verbose) {
 	    verboselog = (FILE *)fopen(dp->verbose,"a");
 	    if (verboselog)
@@ -524,6 +519,10 @@ main(argc, argv)
 	    smtpstatus = EX_OK; /* Already open, we have a new job file */
 
 	  smtpstatus = process(dp, smtpstatus, host, smtpfp, verboselog);
+	  if (verboselog)
+	    fclose(verboselog);
+	  verboselog = NULL;
+
 	  ctlclose(dp);
 	}
 	if (smtpstatus == EX_OK) {
@@ -832,7 +831,7 @@ deliver(dp, smtpfp, verboselog, startrp, endrp)
 	   */
 	  if (smtp_isopen) {
 	    fclose(smtpfp[0]); smtpfp[0] = NULL;
-	    fclose(smtpfp[0]); smtpfp[0] = NULL;
+	    fclose(smtpfp[1]); smtpfp[1] = NULL;
 	    if (logfp)
 	      fprintf(logfp, "%d#\t(closed SMTP channel - appendlet() failure)\n", pid);
 	    smtp_isopen = 0;
@@ -1477,7 +1476,7 @@ smtpconn(host, fparr, verboselog)
 	     * (needed when target of a CNAME has no other RR's)
 	     */
 	    strcpy(hbuf, host);
-	    if ((i = getrr(hbuf, sizeof hbuf, (u_short)T_CNAME),2) == 1) {
+	    if ((i = getrr(hbuf, sizeof hbuf, (u_short)T_CNAME,2)) == 1) {
 	      host = hbuf;
 	    } else if (h_errno == TRY_AGAIN) {
 	      sprintf(remotemsg, "466 (Canonicalize, try later: %s)", host);
@@ -1609,6 +1608,7 @@ makeconn(hp, fdp, verboselog, fparr, mxcnt)
 	int mxcnt;
 {
 	int retval;
+	int fd2;
 
 #ifdef	BIND
 #ifdef	RFC974
@@ -1653,15 +1653,16 @@ makeconn(hp, fdp, verboselog, fparr, mxcnt)
 		int err;
 fdopen_fail:
 		err = errno;
-		if (fparr[0] != NULL) fclose(fparr[0]); fparr[0] = NULL;
-		if (fparr[1] != NULL) fclose(fparr[1]); fparr[1] = NULL;
 		printf("smtp: Failed to fdopen() a socket stream, errno=%d, err='%s' Hmm ??\n",err, strerror(err));
 		fflush(stdout);
 		abort();
+		if (fparr[0] != NULL) fclose(fparr[0]); fparr[0] = NULL;
+		if (fparr[1] != NULL) fclose(fparr[1]); fparr[1] = NULL;
 	      }
-	      if ((fparr[1] = fdopen(*fdp, "w")) == NULL) {
+	      fd2 = dup(*fdp);
+	      if (fd2 < 0 || (fparr[1] = fdopen(fd2, "w")) == NULL)
 		goto fdopen_fail;
-	      }
+
 	      smtp_isopen = 1;
 
 	      if (esmtp_on_banner > 0)
@@ -2302,8 +2303,8 @@ getmxrr(host, mx, maxmx, depth, mxcntp)
 	}
 
 
-	qlen = res_mkquery(QUERY, host, C_IN, T_MX, (char *)NULL, 0, NULL,
-			(char *)&qbuf, sizeof qbuf);
+	qlen = res_mkquery(QUERY, host, C_IN, T_MX, NULL, 0, NULL,
+			   (void*)&qbuf, sizeof qbuf);
 	if (qlen < 0) {
 	  fprintf(stderr, "res_mkquery failed\n");
 	  sprintf(remotemsg,
@@ -2311,7 +2312,7 @@ getmxrr(host, mx, maxmx, depth, mxcntp)
 	  notaryreport(NULL,NULL,remotemsg);
 	  return EX_SOFTWARE;
 	}
-	n = res_send((char *)&qbuf, qlen, (char *)&answer, sizeof answer);
+	n = res_send((void*)&qbuf, qlen, (void*)&answer, sizeof answer);
 	if (n < 0) {
 	  sprintf(remotemsg,
 		  "466 (Internal: res_send failed on host: %s)",host);
@@ -2368,16 +2369,17 @@ getmxrr(host, mx, maxmx, depth, mxcntp)
 	}
 	nmx = 0;
 	cp = (msgdata *)&answer + sizeof(HEADER);
-	for (; qdcount > 0; --qdcount)
+	for (; qdcount > 0; --qdcount) {
 #if	defined(BIND_VER) && (BIND_VER >= 473)
-		cp += dn_skipname(cp, eom) + QFIXEDSZ;
+	  cp += dn_skipname(cp, eom) + QFIXEDSZ;
 #else	/* !defined(BIND_VER) || (BIND_VER < 473) */
-		cp += dn_skip(cp) + QFIXEDSZ;
+	  cp += dn_skip(cp) + QFIXEDSZ;
 #endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
+	}
 	realname[0] = '\0';
 	maxpref = -1;
 	while (--ancount >= 0 && cp < eom && nmx < maxmx-1) {
-	  n = dn_expand((msgdata *)&answer, eom, cp, buf, sizeof buf);
+	  n = dn_expand((msgdata *)&answer, eom, cp, (void*)buf, sizeof buf);
 	  if (n < 0)
 	    break;
 	  cp += n;
@@ -2391,7 +2393,7 @@ getmxrr(host, mx, maxmx, depth, mxcntp)
 	  cp += sizeof(u_short);
 	  if (type == T_CNAME) {
 	    cp += dn_expand((msgdata *)&answer, eom, cp,
-			    realname, sizeof realname);
+			    (void*)realname, sizeof realname);
 	    saw_cname = 1;
 	    continue;
 	  } else if (type != T_MX)  {
@@ -2400,13 +2402,14 @@ getmxrr(host, mx, maxmx, depth, mxcntp)
 	  }
 	  mx[nmx].pref = _getshort(cp);
 	  cp += sizeof(u_short); /* MX preference value */
-	  n = dn_expand((msgdata *)&answer, eom, cp, buf, sizeof buf);
+	  n = dn_expand((msgdata *)&answer, eom, cp, (void*)buf, sizeof buf);
 	  if (n < 0)
 	    break;
 	  cp += n;
 
-	  hent = gethostbyname((char*)buf); /* This resolves CNAME, should it
-					       should not in case of MX server */
+	  hent = gethostbyname((char*)buf); /* This resolves CNAME, it
+					       should not happen in case
+					       of MX server, though..    */
 	  if (!hent) {
 	    *mxcntp += 1;
 	    continue;		/* Well well.. spurious! */
@@ -2416,11 +2419,11 @@ getmxrr(host, mx, maxmx, depth, mxcntp)
 	  if (matchmyaddresses(hent))
 	    maxpref = mx[nmx].pref;
 
-	  if ((mx[nmx].host = (msgdata *)emalloc(strlen(buf)+1)) == NULL) {
+	  if ((mx[nmx].host = (msgdata *)emalloc(strlen((void*)buf)+1)) == NULL) {
 	    fprintf(stderr, "Out of virtual memory!\n");
 	    exit(EX_OSERR);
 	  }
-	  strcpy(mx[nmx].host, buf);
+	  strcpy(mx[nmx].host, (void*)buf);
 	  /* [mea] Canonicalize this target & get A records.. */
 	  ++nmx;
 	}
@@ -2459,7 +2462,7 @@ getmxrr(host, mx, maxmx, depth, mxcntp)
 	    continue;
 	  strcpy(buf, mx[n].host);
 	  /* It is an MX, it CAN'T have CNAME ! */
-	  if (!getrrtype(buf, sizeof buf, T_WKS, 0)) {
+	  if (!getrrtype((void*)buf, sizeof buf, T_WKS, 0)) {
 	    free(mx[n].host);
 	    mx[n].host = NULL;
 	  }
@@ -2497,148 +2500,6 @@ getmxrr(host, mx, maxmx, depth, mxcntp)
 	return EX_OK;
 }
 
-int
-getrr(host, hbsize, rrtype, cnamelevel)	/* getrrtype with completion */
-	char *host;
-	int hbsize;
-	u_short rrtype;
-	int cnamelevel;
-{
-	int rval;
-	char buf[BUFSIZ], **domain;
-
-	if ((rval = getrrtype(host, hbsize, rrtype, cnamelevel)) > 0)
-	  return rval;
-	for (domain = _res.dnsrch; *domain != NULL; domain++) {
-	  sprintf(buf, "%s.%s", host, *domain);
-	  if ((rval = getrrtype(buf, BUFSIZ, rrtype, cnamelevel)) > 0) {
-	    strncpy(host, buf, hbsize<BUFSIZ?hbsize:BUFSIZ);
-	    host[hbsize - 1] = '\0';
-	    return rval;
-	  }
-	}
-	return 0;
-}
-
-int
-getrrtype(host, hbsize, rrtype, cnamelevel)
-	char *host;
-	int hbsize;
-	int cnamelevel;
-	u_short rrtype;
-{
-
-	HEADER *hp;
-	msgdata *eom, *cp;
-	querybuf buf, answer;
-	int qlen, n, ancount, qdcount, ok;
-	u_short type;
-	msgdata nbuf[BUFSIZ];
-	int first;
-
-	qlen = res_mkquery(QUERY, host, C_IN, rrtype, (char *)NULL, 0, NULL,
-			   (char *)&buf, sizeof(buf));
-	if (qlen < 0) {
-	  fprintf(stderr, "res_mkquery failed\n");
-	  h_errno = NO_RECOVERY;
-	  return -2;
-	}
-	n = res_send((char *)&buf, qlen, (char *)&answer, sizeof(answer));
-	if (n < 0) {
-	  h_errno = TRY_AGAIN;
-	  return -1;
-	}
-	eom = (msgdata *)&answer + n;
-	/*
-	 * find first satisfactory answer
-	 */
-	hp = (HEADER *) &answer;
-	ancount = ntohs(hp->ancount);
-	qdcount = ntohs(hp->qdcount);
-	h_errno = 0;
-	/*
-	 * We don't care about errors here, only if we got an answer
-	 */
-	if (ancount == 0) {
-	  if (rrtype == T_CNAME && hp->rcode == NOERROR) {
-	    if (qdcount > 0 && strchr(host, '.') == NULL) {
-	      cp = (msgdata *)&answer + sizeof(HEADER);
-	      if (dn_expand((msgdata *)&answer,
-			    eom, cp, (u_char*)host, hbsize) >= 0) {
-		if (host[0] == '\0') {
-		  host[0] = '.';
-		  host[1] = '\0';
-		}
-	      }
-	    }
-	    return 1;
-	  }
-	  return 0;
-	}
-	cp = (msgdata *)&answer + sizeof(HEADER);
-	for (; qdcount > 0; --qdcount) {
-#if	defined(BIND_VER) && (BIND_VER >= 473)
-	  cp += dn_skipname(cp, eom) + QFIXEDSZ;
-#else	/* !defined(BIND_VER) || (BIND_VER < 473) */
-	  cp += dn_skip(cp) + QFIXEDSZ;
-#endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
-	}
-	first = 1;
-	ok = rrtype != T_WKS;
-	while (--ancount >= 0 && cp < eom) {
-	  if ((n = dn_expand((msgdata *)&answer, eom, cp, nbuf,
-			     sizeof(nbuf))) < 0)
-	    break;
-	  if (first) {
-	    strncpy(host, nbuf, hbsize);
-	    host[hbsize - 1] = '\0';
-	    first = 0;
-	  }
-	  cp += n;
-	  type = _getshort(cp);
-	  cp += sizeof(u_short); /* type */
-	  cp += sizeof(u_short); /* class */
-	  cp += sizeof(u_long);	/* ttl */
-	  n = _getshort(cp);
-	  cp += sizeof(u_short); /* dlen */
-	  if (type != rrtype) {
-	    cp += n;
-	    continue;
-	  }
-	  /*
-	   * Assume that only one rrtype will be found.  More
-	   * than one is undefined.
-	   */
-	  if (type == T_WKS) {
-	    msgdata *nextcp = cp + n;
-	    /* If we have seen a WKS, it had better have SMTP,
-	     * however in absence of a WKS, assume SMTP.
-	     */
-	    if (n < sizeof (u_long) + 1) {
-	      cp = nextcp;	/* ? */
-	      continue;
-	    }
-	    ok = 0;
-	    cp += sizeof(u_long); /* skip IP address */
-	    if (*cp++ == IPPROTO_TCP) {	/* check protocol */
-	      if (cp + (IPPORT_SMTP/8) < nextcp
-		  && (*(cp+(IPPORT_SMTP/8))
-		      & (0x80>>IPPORT_SMTP%8)))
-		return 1;
-	    }
-	    cp = nextcp;
-	    continue;
-	  }
-	  /* Special processing on T_CNAME ??? */
-	  if ((n = dn_expand((msgdata *)&answer, eom, cp, nbuf,
-			     sizeof(nbuf))) < 0)
-	    break;
-	  strncpy(host, nbuf, hbsize);
-	  host[hbsize - 1] = '\0';
-	  return ok;
-	}
-	return 0;
-}
 
 /*
  * This is the callback function for ctlopen.  It should return 0 to reject
